"""Pulsar phase prediction."""

import operator
import functools
from dataclasses import dataclass, asdict
import numpy as np
from numpy.polynomial import Polynomial
from astropy import units as u
from astropy.time import Time
from astropy.table import QTable
import pulsarbat as pb

__all__ = ["PolycoEntry", "PhasePredictor"]


@dataclass
class PolycoEntry:
    """Entry for polynomial-based pulsar phase prediction.

    Parameters
    ----------
    psr : str
        Pulsar name.
    obs : str
        Observatory code.
    freq : Quantity
        Observing frequency.
    tmid : Time
        Timestamp at midpoint of span.
    span : Quantity
        Length of span.
    rphase : int
        Integer part of reference phase
    poly : numpy.polynomial.Polynomial
        Phase prediction polynomial (domain in units of seconds).
    """

    psr: str
    obs: str
    freq: u.Quantity
    tmid: Time
    span: u.Quantity
    rphase: int
    poly: Polynomial


class PhasePredictor(QTable):
    """Pulsar phase predictor.

    Parameters
    ----------
    data : sequence of PhasePredictorEntry
        A list of phase prediction entries. All entries must have the
        same 'psr', 'obs', 'freq', and 'span' values.
    """

    _descriptions = {
        "psr": "Pulsar name",
        "obs": "Observatory code",
        "freq": "Observing frequency",
        "tmid": "Time at midpoint of span",
        "span": "Length of span",
        "rphase": "Integer part of reference phase",
        "poly": "Phase prediction polynomial (domain in seconds)",
    }

    def __init__(self, data=None, *args, **kwargs):
        try:
            data = [asdict(x) for x in data]
            data = sorted(data, key=operator.itemgetter("tmid"))
        except (TypeError, KeyError):
            pass
        else:
            kwargs.setdefault("descriptions", self._descriptions)

        super().__init__(data, *args, **kwargs)

        for k in ["psr", "obs", "freq", "span"]:
            if k in self.colnames and len(np.unique(self[k])) > 1:
                d = self._descriptions[k]
                raise ValueError(f"All entries must have the same '{k}' ({d}).")

        self._intervals = None

    @property
    def intervals(self):
        """Intervals where the predictor is valid and should be used."""
        if self._intervals is None:
            tstart = self["tmid"] - self["span"] / 2
            tstop = self["tmid"] + self["span"] / 2
            intervals = sorted(zip(tstart, tstop), key=lambda x: x[1])

            merged = []
            start, end = intervals.pop()
            while intervals:
                next_start, next_end = intervals.pop()
                if next_end >= start or start.isclose(next_end, 1 * u.ms):
                    start = min(start, next_start)
                else:
                    merged.append([start, end])
                    start, end = next_start, next_end
            merged.append([start, end])
            self._intervals = tuple(reversed(merged))

        return self._intervals

    def _get_index_and_dt(self, times):
        """Check if timestamps are within predictor range."""
        check = ((a <= times) & (times <= b) for a, b in self.intervals)
        check = functools.reduce(operator.or_, check)

        if not np.all(check):
            raise ValueError("Some timestamps outside predictor range!")

        span_ends = self["tmid"] + self["span"] / 2
        index = np.searchsorted(span_ends.mjd, times.mjd)
        dt = (times - self["tmid"][index]).to_value(u.s)
        return index, dt

    def __call__(self, times):
        """Predict pulse phase at given times.

        Parameters
        ----------
        times : Time
            Timestamp(s) for which phases are to be predicted.

        Returns
        -------
        phase : pulsarbat.Phase
            Predicted phase for given timestamps.
        """
        index, dt = self._get_index_and_dt(times)

        if times.isscalar:
            ph1 = self["rphase"][index]
            ph2 = self["poly"][index](dt)
        else:
            ph1 = np.zeros(times.shape, dtype=np.int64)
            ph2 = np.zeros(times.shape, dtype=np.float64)
            for i in np.unique(index):
                s = index == i
                ph1[s] = self["rphase"][i]
                ph2[s] = self["poly"][i](dt[s])

        return pb.Phase(ph1, ph2)

    def phasepol(self, t0):
        """Phase prediction polynomial centered at given timestamp."""
        if not t0.isscalar:
            raise ValueError("Timestamp must be a scalar.")

        index, dt = self._get_index_and_dt(t0)
        rphase = self["rphase"][index]
        polynomial = self["poly"][index].copy()
        polynomial.domain -= dt
        a = int(polynomial(0) // 1)

        return (polynomial - a).convert(), pb.Phase(rphase + a)

    def f0(self, times, n=0):
        """Predict rotation frequency or its derivatives for given times."""
        index, dt = self._get_index_and_dt(times)
        unit = u.cycle / u.s ** (n + 1)

        if times.isscalar:
            f = self["poly"][index].deriv(n + 1)(dt)
        else:
            f = np.zeros(times.shape, dtype=np.float64)
            for i in np.unique(index):
                s = index == i
                f[s] = self["poly"][i].deriv(n + 1)(dt[s])
        return f * unit

    @classmethod
    def from_polyco(cls, path):
        """Create a PhasePredictor instance by reading tempo1-style polycos.

        Parameters
        ----------
        path : path-like or file-like
            Either a path to a polyco file or a file-like object that
            supports the ``readline()`` method.

        Examples
        --------
        From reading polyco files,

        >>> predictor = pb.PhasePredictor.from_polyco("polyco.dat")

        Alternatively, a text stream can be used (such as ``io.StringIO``),

        >>> import io
        >>> str_io = io.StringIO(polyco_text)
        >>> predictor = pb.PhasePredictor.from_polyco(str_io)

        Notes
        -----
        The format of a polyco entry (a file may have more than one) is

        .. code-block:: text

            ====  =======   ============================================
            Line  Columns   Item
            ====  =======   ============================================
            1        1-10   Pulsar Name
                    12-20   Date (dd-mmm-yy)
                    21-31   UTC (hhmmss.ss)
                    32-51   TMID (MJD)
                    52-72   Dispersion Measure (pc / cm^3)
                    74-79   Doppler shift due to earth motion (10^-4)
                    80-86   Log_10 of fit rms residual in periods
            2        1-20   Reference Phase (RPHASE)
                    22-38   Reference rotation frequency (F0)
                    39-43   Observatory number
                    44-49   Data span (minutes)
                    50-54   Number of coefficients (NCOEFF)
                    55-64   Observing frequency (MHz)
                    65-71   (Optional) Binary orbit phase
                    72-80   (Optional) Orbital frequency (1/day)
            3*       1-25   Coefficient 1 (COEFF(1))
                    26-50   Coefficient 2 (COEFF(2))
                    51-75   Coefficient 3 (COEFF(3))
            ====  =======   ============================================
            * Subsequent lines have three coefficients each, up to NCOEFF

        The pulse phase and frequency at time T (in MJD) are then calculated as::

            DT = (T-TMID)*1440
            PHASE = RPHASE + DT*60*F0 + COEFF(1) + DT*COEFF(2) + DT^2*COEFF(3) + ....
            FREQ(Hz) = F0 + (1/60)*(COEFF(2) + 2*DT*COEFF(3) + 3*DT^2*COEFF(4) + ....)

        Example tempo2 call to produce one:

        .. code-block:: text

            tempo2 -tempo1 -f pulsar.par
                -polyco "56499 56500 90 12 12 ao 1400"
                         |-- MJD start
                               |-- MJD end
                                     |-- length of span (in minutes)
                                        |-- Number of polynomial coefficients
                                           |-- Max Hour Angle (12 is continuous)
                                              |-- Observatory code
                                                 |-- Frequency in MHz

        References
        ----------
        http://tempo.sourceforge.net/ref_man_sections/tz-polyco.txt
        https://bitbucket.org/psrsoft/tempo2/
        """

        f = path if hasattr(path, "readline") else open(path, "r")
        d2e = str.maketrans("Dd", "ee")

        table = []
        with f:
            while (line := f.readline()) :
                psr, _, _, mjd_mid, dm, *_ = line.split()
                rphase, f0, obs, span, ncoeff, freq, *_ = f.readline().split()
                r_int, _, r_frac = rphase.partition(".")

                coeffs = []
                for _ in range(-(int(ncoeff) // -3)):
                    coeffs += f.readline().translate(d2e).split()

                coeffs = np.array(coeffs, dtype=np.float64)
                coeffs[0] += float("0." + r_frac)
                coeffs[1] += float(f0) * 60

                entry = PolycoEntry(
                    psr=psr,
                    obs=obs,
                    freq=float(freq) * u.MHz,
                    tmid=Time(mjd_mid, format="mjd", precision=9),
                    span=int(span) * u.min,
                    rphase=np.int64("0" + r_int),
                    poly=Polynomial(coeffs, domain=[-60, +60]).convert(),
                )

                table.append(entry)

        return cls(table)



"""Base module for readers.

Readers are used to read data into `~Signal` objects. A reader `r`
should expose the following user-facing attributes/methods:

  * `len(r)` -- Length of data in number of samples.
  * `r.shape` -- Shape of data.
  * `r.sample_shape` -- Shape of a sample.
  * `r.ndim` -- Number of dimensions.
  * `r.dtype` -- Data-type of signal data.
  * `r.sample_rate` -- Number of samples per unit time.
  * `r.dt` -- Sample spacing in time units.
  * `r.time_length` -- Length of signal data in time units.
  * `r.start_time` -- Timestamp at first sample (or None).
  * `r.stop_time` -- Timestamp after last sample (or None).
  * `r.read(offset, n, **kwargs)` -- Read `n` samples from given offset.
  * `r.offset_at(t)` -- Offset at given timestamp or Quantity.
  * `r.time_at(offset)` -- Timestamp at given offset (or None).
  * `r.contains(t)` -- Whether time(s) are within the bounds of the data.
  * `t in r` -- Whether times is within the bounds of the data.

The following keyword arguments should be accepted the `read()` method:
  * `use_dask` -- boolean, whether to use Dask arrays.
  * `chunks` -- Chunk sizes if using dask arrays.
"""

import operator
import numpy as np
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb

__all__ = [
    "BaseReader",
]


class OutOfBoundsError(EOFError):
    """Raised when stream position is out of bounds."""

    pass


class BaseReader:
    """Base class for readers.

    Subclasses must either implement the ``_read_array()`` method if using
    the default ``.read()`` implementation, or implement their own
    ``.read()`` method.

    Parameters
    ----------
    shape : tuple of int
        Signal shape.
    dtype : dtype
        Data-type of signal data.
    signal_type : subclass of .Signal, default: .Signal
        Type of signal that will be returned by ``read()``.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        Timestamp at first sample of signal data. Default is None.
    **signal_kwargs
        Additional kwargs to pass on to ``signal_type`` when creating a
        :py:class:`.Signal` object.
    """

    def __init__(
        self,
        /,
        *,
        shape,
        dtype,
        signal_type=pb.Signal,
        sample_rate,
        start_time=None,
        **signal_kwargs,
    ):

        if not issubclass(signal_type, pb.Signal):
            raise ValueError("Bad signal_type. Must be Signal or subclass.")

        self._signal_type = signal_type
        self._signal_kwargs = signal_kwargs
        for key in signal_kwargs:
            setattr(self, key, signal_kwargs[key])

        self._dtype = np.dtype(dtype)
        self._shape = tuple(operator.index(a) for a in shape)

        if not self.ndim:
            raise ValueError("Invalid shape.")

        self.sample_rate = sample_rate
        self.start_time = start_time

        # Read 0 samples now to catch potential errors earlier
        z = self.read(0, 0)

        # Make sure shape and dtype are consistent
        if z.shape != (0,) + self.sample_shape:
            raise ValueError("Provided shape does not match output shape!")

        if z.dtype != self.dtype:
            raise ValueError("Provided dtype does not match output dtype!")

    def _attr_repr(self):
        st = "N/A" if self.start_time is None else self.start_time.isot
        return (
            f"Start time: {st}\n"
            f"Sample rate: {self.sample_rate}\n"
            f"Time length: {self.time_length}\n"
        )

    def __str__(self):
        signature = f"{self.__class__.__name__} @ {hex(id(self))}"
        info = f"shape={self.shape}, dtype={self.dtype}"
        s = f"{signature}\n{'-' * len(signature)}\n"
        s += f"Data Container: {self._signal_type.__name__}<{info}>\n"
        s += self._attr_repr()
        return s.strip()

    def __repr__(self):
        info = f"shape={self.shape}, dtype={self.dtype}"
        sig = f"{self._signal_type.__name__}({info})"
        s = f"{self.__class__.__name__}<{sig}> @ {hex(id(self))}"
        return s

    def __dir__(self):
        members = set(object.__dir__(self))
        members.update(self._signal_kwargs)
        return sorted(members)

    def __len__(self):
        """Length of signal data in number of samples."""
        return self.shape[0]

    @property
    def shape(self):
        """Shape of data."""
        return self._shape

    @property
    def sample_shape(self):
        """Shape of a sample."""
        return self.shape[1:]

    @property
    def ndim(self):
        """Number of dimensions in data."""
        return len(self.shape)

    @property
    def dtype(self):
        """Data-type of data."""
        return self._dtype

    @property
    def sample_rate(self):
        """Sample rate of the signal data."""
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, sample_rate):
        try:
            temp = sample_rate.to(u.Hz)
            assert temp.isscalar and temp > 0
        except Exception:
            raise ValueError(
                "Invalid sample_rate. Must be a positive scalar "
                "Quantity with units of Hz or equivalent."
            )
        else:
            self._sample_rate = sample_rate

    @property
    def start_time(self):
        """Timestamp at first sample."""
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        try:
            temp = None
            if start_time is not None:
                temp = Time(start_time, format="isot", precision=9)
                assert temp.isscalar
        except Exception:
            raise ValueError(
                "Invalid start_time. Must be a scalar astropy " "Time object."
            )
        else:
            self._start_time = temp

    @property
    def stop_time(self):
        """Timestamp after last sample."""
        return self.time_at(len(self))

    @property
    def dt(self):
        """Sample spacing (1 / sample_rate)."""
        return (1 / self.sample_rate).to(u.s)

    @property
    def time_length(self):
        """Length of signal in time units."""
        return (len(self) / self.sample_rate).to(u.s)

    def contains(self, t, /):
        """Whether time(s) are within the bounds of the signal."""
        if self.start_time is None:
            return np.zeros(t.shape, bool) if t.shape else False

        t0, t1 = self.start_time, self.stop_time
        edge = ~Time.isclose(t, t1) | Time.isclose(t, t0)
        return edge & (t0 <= t) & (t < t1)

    def __contains__(self, t):
        """Whether time is within the bounds of the signal."""
        return self.contains(t)

    def offset_at(self, t, /):
        """Returns nearest integer offset at given time.

        Parameters
        ----------
        t : Quantity or Time
            Can be an absolute time as an astropy Time object, or an
            astropy Quantity in time units relative to the start.

        Returns
        -------
        offset : int
            The nearest integer position (in number of samples).
        """
        try:
            t = t - self.start_time
        except Exception:
            pass

        offset = int((t * self.sample_rate).to(u.one).round())

        if offset < 0 or offset > len(self):
            raise OutOfBoundsError("Given time is out of bounds!")

        return offset

    def time_at(self, offset, /, unit=None):
        """Returns time at given offset.

        Parameters
        ----------
        offset : int
            Position in number of samples.
        unit : Unit, optional
            Desired unit of returned value (as an astropy unit).
            By default (None), the absolute timestamp is returned.

        Returns
        -------
        t : Quantity or Time
            Time relative to the start as an astropy Quantity if `unit`
            is provided, otherwise an absolute time as an astropy Time
            object.
        """
        if unit is not None:
            return (offset / self.sample_rate).to(unit)

        if self.start_time is None:
            return None

        return self.start_time + (offset / self.sample_rate)

    def _read_array(self, offset, n, /):
        """Read n samples from given offset into numpy array."""
        return NotImplemented

    def _read_data(self, offset, n, /, use_dask=False, **kwargs):
        """Read n samples from current read position into array-like."""
        if use_dask:
            import dask
            import dask.array as da

            delayed_read = dask.delayed(self._read_array, pure=True)

            _out_shape = (n,) + self.sample_shape
            z = da.from_delayed(delayed_read(offset, n, **kwargs),
                                dtype=self.dtype, shape=_out_shape)

            default_chunks = (-1,) + ("auto",) * len(self.sample_shape)
            z = z.rechunk(kwargs.get("chunks", default_chunks))
        else:
            z = self._read_array(offset, n, **kwargs)

        return z

    def read(self, offset, n, /, **kwargs):
        """Read `n` samples from given offset.

        Parameters
        ----------
        offset : int
            Position to read from. Must be non-negative.
        n : int
            Number of samples to read. Must be non-negative.
        **kwargs
            Currently supported keyword arguments:

              * ``use_dask`` -- Whether to use dask arrays.
              * ``chunks`` -- Chunk sizes if using dask arrays. By default,
                there is no chunking along the zeroth dimension.

        Returns
        -------
        out : Signal
            Signal of length ``n`` containing data that was read.
        """
        if (offset := operator.index(offset)) < 0:
            raise ValueError("offset must be a non-negative int.")

        if (n := operator.index(n)) < 0:
            raise ValueError("n must be a non-negative int.")

        if offset + n > len(self):
            raise OutOfBoundsError("Cannot read beyond end of stream")

        return self._signal_type(
            self._read_data(offset, n, **kwargs),
            sample_rate=self.sample_rate,
            start_time=self.time_at(offset),
            **self._signal_kwargs,
        )

    def dask_read(self, offset, n, /, **kwargs):
        """Read `n` samples from given offset using Dask arrays.

        A convenience method equivalent to the :py:meth:`.read()` method with
        ``use_dask=True``.

        Parameters
        ----------
        offset : int
            Position to read from. Must be non-negative.
        n : int
            Number of samples to read. Must be non-negative.
        **kwargs
            Additional keyword arguments. Currently supported:
              * ``chunks`` -- Chunk sizes if using dask arrays.
                              By default, there is no chunking along
                              the zeroth dimension.

        Returns
        -------
        out : Signal
            Signal of length ``n`` containing data that was read as a
            Dask array.
        """
        return self.read(offset, n, use_dask=True, **kwargs)



"""Core module defining Signal classes."""

import operator
import inspect
import pprint
import numpy as np
import astropy.units as u
from astropy.time import Time
import dask.array


__all__ = [
    "Signal",
    "RadioSignal",
    "IntensitySignal",
    "FullStokesSignal",
    "BasebandSignal",
    "DualPolarizationSignal",
]


class InvalidSignalError(ValueError):
    """Used to catch invalid signals."""

    pass


class Signal(np.lib.mixins.NDArrayOperatorsMixin):
    """Base class for all signals.

    A signal is sufficiently described by an array of samples (``z``),
    and a sampling rate (``sample_rate``). Optionally, a ``start_time`` can
    be provided to specify the time stamp at the first sample of the
    signal.

    The signal data (``z``) must have at least 1 dimension where the
    zeroth axis (``axis=0``) refers to time. That is, ``z[i]`` is the
    ``i``-th sample of the signal, and the sample shape (``z[i].shape``)
    can be arbitrary.

    Parameters
    ----------
    z : array-like
        The signal data. Must be at least 1-dimensional with shape
        ``(nsample, ...)``, and must have non-zero size.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        The start time of the signal (that is, the time at the first
        sample of the signal).
    meta : dict, optional
        Any metadata that the user might want to attach to the signal.
    """

    _req_dtype = ()
    _req_shape = (None,)
    _axes_labels = {"time": 0}

    def __init__(self, z, /, *, sample_rate, start_time=None, meta=None):
        min_ndim = len(self._req_shape)

        if z.ndim < min_ndim:
            raise InvalidSignalError(
                f"Expected signal with at least {min_ndim} dimension(s), "
                f"got signal with {z.ndim} dimension(s) instead."
            )

        zipped = zip(z.shape[:min_ndim], self._req_shape)
        if not all(x == (y or x) for x, y in zipped):
            raise InvalidSignalError(
                f"Signal has invalid shape. Expected {self._req_shape}, "
                f"got {z.shape} instead."
            )

        if np.prod(z.shape[1:]) == 0:
            raise InvalidSignalError("Sample shape must have non-zero size!")

        _temp = None
        if z.dtype in self._req_dtype or not self._req_dtype:
            _temp = z
        else:
            for d in self._req_dtype:
                try:
                    _temp = z.astype(self._req_dtype[0], casting="safe")
                except TypeError:
                    pass
                else:
                    break

        if _temp is None:
            err = f"Expected {self._req_dtype}, got {z.dtype}."
            raise InvalidSignalError(f"Invalid dtype. {err}")

        self._data = _temp
        self.sample_rate = sample_rate
        self.start_time = start_time
        self.meta = meta

    def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):
        if method != "__call__" or ufunc == np.matmul:
            return NotImplemented

        in_arr = tuple((i.data if isinstance(i, Signal) else i) for i in inputs)

        if out is None:
            out = (None,) * ufunc.nout

        out_arr = tuple((i.data if isinstance(i, Signal) else i) for i in out)

        results = ufunc(*in_arr, out=out_arr, **kwargs)

        if results is NotImplemented:
            return NotImplemented

        if ufunc.nout == 1:
            results = (results,)

        results = tuple(
            (type(self).like(self, a) if b is None else b) for a, b in zip(results, out)
        )

        return results[0] if len(results) == 1 else results

    def _attr_repr(self):
        st = "N/A" if self.start_time is None else self.start_time.isot
        return (
            f"Sample rate: {self.sample_rate}\n"
            f"Time length: {self.time_length}\n"
            f"Start time: {st}\n"
        )

    def __str__(self):
        signature = f"{self.__class__.__name__} @ {hex(id(self))}"
        c = type(self.data)
        s = f"{signature}\n{'-' * len(signature)}\n"
        s += f"Data Container: {c.__module__}.{c.__name__}"
        s += f"<shape={self.shape}, dtype={self.dtype}>\n"
        s += self._attr_repr()
        if self.meta is not None:
            s += "\nMeta\n----\n"
            s += pprint.pformat(self.meta, sort_dicts=False, depth=2)
        return s

    def __repr__(self):
        info = f"shape={self.shape}, dtype={self.dtype}"
        s = f"pulsarbat.{self.__class__.__name__}<{info}> @ {hex(id(self))}"
        return s

    def __len__(self):
        return len(self.data)

    def __array__(self):
        return np.asanyarray(self.data)

    def _time_slice(self, index):
        s = slice(*index.indices(self.shape[0]))
        assert s.step > 0, "Time axis slicing does not support negative step"

        kw = dict()
        if s.step > 1:
            kw["sample_rate"] = self.sample_rate / s.step
        if self.start_time is not None:
            kw["start_time"] = self.start_time + s.start / self.sample_rate
        return kw

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        if not all(isinstance(a, slice) for a in index[:1]):
            err = "Only supports slicing on time axis."
            raise IndexError(err)

        kw = dict()
        kw.update(self._time_slice(index[0]))
        return type(self).like(self, self.data[index], **kw)

    def get_axis(self, axis):
        """Returns axis from an integer or axis label."""
        try:
            axis = operator.index(axis)
        except TypeError:
            axis = self.axes_labels.get(axis, None)

        if (axis is None) or (axis < -self.ndim) or (self.ndim <= axis):
            raise ValueError("Invalid axis.")

        return axis

    @property
    def axes_labels(self):
        """Dictionary of axes labels."""
        return self._axes_labels

    @property
    def meta(self):
        """Signal metadata."""
        return self._meta

    @meta.setter
    def meta(self, meta):
        try:
            self._meta = None if meta is None else dict(meta)
        except Exception:
            raise ValueError("meta must be a dict.")

    @property
    def data(self):
        """The signal data."""
        return self._data

    @property
    def shape(self):
        """Shape of the signal."""
        return self.data.shape

    @property
    def sample_shape(self):
        """Shape of a sample."""
        return self.shape[1:]

    @property
    def ndim(self):
        """Number of dimensions in data."""
        return self.data.ndim

    @property
    def dtype(self):
        """Data type of the signal."""
        return self.data.dtype

    @property
    def sample_rate(self):
        """Sample rate of the signal."""
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, sample_rate):
        try:
            temp = sample_rate.to(u.Hz)
            assert temp.isscalar and temp > 0
        except Exception:
            raise ValueError(
                "Invalid sample_rate. Must be a positive scalar "
                "Quantity with units of Hz or equivalent."
            )
        else:
            self._sample_rate = sample_rate

    @property
    def dt(self):
        """Sample spacing of the signal (1 / sample_rate)."""
        return (1 / self.sample_rate).to(u.s)

    @property
    def time_length(self):
        """Length of signal in time units."""
        return (len(self) / self.sample_rate).to(u.s)

    @property
    def start_time(self):
        """Start time of the signal (Time at first sample)."""
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        try:
            temp = None
            if start_time is not None:
                temp = Time(start_time, format="isot", precision=9)
                assert temp.isscalar
        except Exception:
            err = "Invalid start_time. Must be a scalar astropy Time object."
            raise ValueError(err)
        else:
            self._start_time = temp

    @property
    def stop_time(self):
        """Stop time of the signal (Time at sample after the last sample)."""
        if self.start_time is None:
            return None
        return self.start_time + self.time_length

    def contains(self, t, /):
        """Whether time(s) are within the bounds of the signal."""
        if self.start_time is None:
            return np.zeros(t.shape, bool) if t.shape else False

        t0, t1 = self.start_time, self.stop_time
        edge = ~Time.isclose(t, t1) | Time.isclose(t, t0)
        return edge & (t0 <= t) & (t < t1)

    def __contains__(self, t):
        """Whether time is within the bounds of the signal."""
        return self.contains(t)

    def compute(self, **kwargs):
        """Returns signal with computed data.

        Has no effect unless data is stored as a Dask Array. ``kwargs`` are
        passed on to the :py:func:`dask.compute` function.
        """
        if isinstance(self.data, dask.array.Array):
            x = self.data.compute(**kwargs)
        else:
            x = np.asarray(self.data)

        return type(self).like(self, x)

    def persist(self, **kwargs):
        """Returns signal with data persisted in memory.

        Has no effect unless data is stored as a Dask Array. ``kwargs`` are
        passed on to the :py:func:`dask.persist` function.
        """
        if isinstance(self.data, dask.array.Array):
            x = self.data.persist(**kwargs)
        else:
            x = np.asarray(self.data)

        return type(self).like(self, x)

    def to_dask_array(self):
        """Returns signal with data as a Dask array.

        Uses :py:func:`dask.array.asanyarray` to convert signal data
        to a Dask array.
        """
        return type(self).like(self, dask.array.asanyarray(self.data))

    def rechunk(self, chunks=None, **kwargs):
        """Rechunks signal data if stored as a Dask array.

        If the underlying data is not a Dask array, it is converted to one first.
        By default, there is no chunking along the first dimension and other
        dimensions are automatically chunked based on the default chunk size set in
        Dask's configuration. ``kwargs`` are passed along to
        :py:func:`dask.array.rechunk`.
        """
        if chunks is None:
            chunks = (-1,) + ("auto",) * (self.ndim - 1)

        x = dask.array.asanyarray(self.data)
        return type(self).like(self, x.rechunk(chunks, **kwargs))

    @classmethod
    def like(cls, obj, z=None, /, **kwargs):
        """Creates a signal object "like" another signal object.

        This classmethod inspects the class signature and creates a
        signal object like the given reference object. The signal data
        (as an array) must always be provided as well as any other
        arguments that are intended to be different than the reference
        object. All arguments required by the class signature that are
        not provided are acquired from attributes of the same name in
        the reference object.

        Parameters
        ----------
        obj : Signal
            The reference signal object.
        z : array-like
            The signal data.
        **kwargs
            Additional keyword arguments to pass on to the target class.
        """
        sig = inspect.signature(cls)

        for k, v in sig.parameters.items():
            if v.kind is not v.POSITIONAL_ONLY and k not in kwargs:
                if hasattr(obj, k):
                    kwargs[k] = getattr(obj, k)
                elif v.default is v.empty:
                    raise ValueError(f"Missing required keyword argument: {k}")

        if z is None:
            z = obj.data
        return cls(z, **kwargs)


class RadioSignal(Signal):
    """Class for heterodyned radio signals.

    A radio signal is often heterodyned, so a center frequency
    (``center_freq``) and a channel bandwidth (``chan_bw``) must be
    provided to determine the frequencies of each channel. Optionally,
    ``freq_align`` is provided to indicate how the channel frequencies
    are positioned. Usually, the signal data here would be the result
    of some sort of filterbank.

    The signal data (``z``) must have at least 2 dimensions where the
    first two dimensions refer to time (``axis=0``) and frequency
    (``axis=1``), respectively.

    Parameters
    ----------
    z : array-like
        The signal data. Must be at least 2-dimensional with shape
        ``(nsample, nchan, ...)``.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : Quantity
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : Quantity
        The bandwidth of a channel. The total bandwidth is ``chan_bw *
        nchan``. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, default: 'center'
        The alignment of channels relative to the ``center_freq``. Default
        is 'center' (as with odd-length complex DFTs). 'bottom' and
        'top' only have an effect when ``nchan`` is even.
    meta : dict, optional
        Any metadata that the user might want to attach to the signal.

    Notes
    -----
    The input signal array must have shape ``(nsample, nchan, ...)``,
    where ``nsample`` is the number of time samples, and ``nchan`` is the
    number of frequency channels.

    The channels must be adjacent in frequency and of equal channel
    bandwidth, such that the frequency of a channel ``i`` is given by,::

        freq_i = center_freq + chan_bw * (i + a - nchan/2)

    where ``i`` is in ``[0, ..., nchan - 1]``, ``nchan`` is the number of
    channels (``z.shape[1]``), ``chan_bw`` is the bandwidth of a single
    channel, and ``a`` is in ``{0, 0.5, 1}`` depending on the value of
    ``freq_align``.

    An even-length complex-valued DFT (as implemented in `~numpy.fft`)
    would have ``freq_align = 'bottom'`` with a center frequency of 0,
    whereas an odd-length DFT would have ``freq_align = 'center'`` due to
    symmetry. ``freq_align = 'top'`` is used in cases where the DFT was
    conducted on the lower sideband of the signal instead.

    When ``nchan`` is odd, the ``freq_align`` attribute will internally be
    fixed to 'center' since the channels must be necessarily centered
    on the ``center_freq``.
    """

    _req_shape = (None, None)
    _axes_labels = {"time": 0, "freq": 1}

    def __init__(
        self,
        z,
        /,
        *,
        sample_rate,
        start_time=None,
        center_freq,
        chan_bw,
        freq_align="center",
        meta=None,
    ):

        super().__init__(z, sample_rate=sample_rate, start_time=start_time, meta=meta)

        self.center_freq = center_freq
        self.chan_bw = chan_bw
        self.freq_align = freq_align

    def _attr_repr(self):
        s = super()._attr_repr()
        s += f"Channel Bandwidth: {self.chan_bw}\n"
        s += f"Total Bandwidth: {self.bandwidth}\n"
        s += f"Center Frequency: {self.center_freq}\n"
        return s

    def _freq_slice(self, index):
        s = slice(*index.indices(self.shape[1]))
        assert s.step == 1, "Does not support slice step for frequency axis"
        assert s.stop > s.start, "Empty frequency slice!"
        f = self.channel_freqs[s]
        return {"center_freq": (f[0] + f[-1]) / 2, "freq_align": "center"}

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        if not all(isinstance(a, slice) for a in index[:2]):
            err = "Only supports slicing on time and frequency axes."
            raise IndexError(err)

        kw = dict()
        kw.update(self._time_slice(index[0]))
        if len(index) > 1:
            kw.update(self._freq_slice(index[1]))
        return type(self).like(self, self.data[index], **kw)

    @property
    def nchan(self):
        """Number of frequency channels."""
        return self.shape[1]

    @property
    def center_freq(self):
        """Center frequency."""
        return self._center_freq

    @center_freq.setter
    def center_freq(self, center_freq):
        try:
            temp = center_freq.to(u.Hz)
            assert temp.isscalar
        except Exception:
            raise ValueError(
                "Invalid center_freq. Must be a scalar "
                "Quantity with units of Hz or equivalent."
            )
        else:
            self._center_freq = center_freq

    @property
    def bandwidth(self):
        """Total bandwidth."""
        return self.chan_bw * self.nchan

    @property
    def chan_bw(self):
        """Channel bandwidth."""
        return self._chan_bw

    @chan_bw.setter
    def chan_bw(self, chan_bw):
        try:
            temp = chan_bw.to(u.Hz)
            assert temp.isscalar and temp > 0
        except Exception:
            raise ValueError(
                "Invalid chan_bw. Must be a positive scalar "
                "Quantity with units of Hz or equivalent."
            )
        else:
            self._chan_bw = chan_bw

    @property
    def max_freq(self):
        """Frequency at top of the band."""
        return self.center_freq + self.bandwidth / 2

    @property
    def min_freq(self):
        """Frequency at bottom of the band."""
        return self.center_freq - self.bandwidth / 2

    @property
    def freq_align(self):
        """Alignment of channel frequencies."""
        return self._freq_align

    @freq_align.setter
    def freq_align(self, freq_align):
        if freq_align in {"bottom", "center", "top"}:
            self._freq_align = "center" if self.nchan % 2 else freq_align
        else:
            choices = "{'bottom', 'center', 'top'}"
            raise ValueError(f"Invalid freq_align. Expected: {choices}")

    @property
    def channel_freqs(self):
        """Returns a list of frequencies corresponding to all channels."""
        _align = {"bottom": 0, "center": 0.5, "top": 1}[self.freq_align]
        chan_ids = np.arange(self.nchan) + _align - self.nchan / 2
        return self.center_freq + self.chan_bw * chan_ids


class IntensitySignal(RadioSignal):
    """Class for intensity signals.

    See the documentation for :py:class:`RadioSignal` for more details.

    Parameters
    ----------
    z : array-like
        The signal data. Must be at least 3-dimensional with shape
        ``(nsample, nchan, npol, ...)``.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : Quantity
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : Quantity
        The bandwidth of a channel. The total bandwidth is ``chan_bw *
        nchan``. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, default: 'center'
        The alignment of channels relative to the ``center_freq``. Default
        is 'center' (as with odd-length complex DFTs). 'bottom' and
        'top' only have an effect when ``nchan`` is even.
    meta : dict, optional
        Any metadata that the user might want to attach to the signal.

    See Also
    --------
    Signal
    RadioSignal
    """

    _req_dtype = (np.float64, np.float32)


class FullStokesSignal(IntensitySignal):
    """Class for full Stokes (I, Q, U, V) signals.

    See the documentation for :py:class:`RadioSignal` and
    :py:class:`IntensitySignal` for more details. We use the PSR/IEEE
    convention for Stokes parameters. Stokes V is positive for
    left-handed circular polarization.

    Parameters
    ----------
    z : array-like
        The signal data. Must be at least 3-dimensional with shape
        ``(nsample, nchan, nstokes, ...)`` where ``nstokes = 4``.
        The order of the Stokes components is ``[I, Q, U, V]``.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : Quantity
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    chan_bw : Quantity
        The bandwidth of a channel. The total bandwidth is ``chan_bw *
        nchan``. Must be in units of frequency.
    freq_align : {'bottom', 'center', 'top'}, default: 'center'
        The alignment of channels relative to the ``center_freq``. Default
        is 'center' (as with odd-length complex DFTs). 'bottom' and
        'top' only have an effect when ``nchan`` is even.
    meta : dict, optional
        Any metadata that the user might want to attach to the signal.

    See Also
    --------
    Signal
    RadioSignal
    IntensitySignal

    References
    ----------
    Wikipedia, "Stokes Parameters",
    https://en.wikipedia.org/wiki/Stokes_parameters

    van Straten et al. (2009),
    https://doi.org/10.1071/AS09084
    """

    _req_shape = (None, None, 4)
    _axes_labels = {"time": 0, "freq": 1, "pol": 2}
    _stokes_ids = {"I": 0, "Q": 1, "U": 2, "V": 3}

    def __getitem__(self, key):
        if isinstance(key, str):
            index = self._stokes_ids.get(key)
            if index is None:
                err = "Invalid key. Should be in {'I', 'Q', 'U', 'V'}."
                raise KeyError(err)
            else:
                axis = self.get_axis("pol")
                x = np.take(self.data, index, axis=axis)
                return IntensitySignal.like(self, x)
        else:
            return super().__getitem__(key)

    @property
    def stokesI(self):
        return self["I"]

    @property
    def stokesQ(self):
        return self["Q"]

    @property
    def stokesU(self):
        return self["U"]

    @property
    def stokesV(self):
        return self["V"]


class BasebandSignal(RadioSignal):
    """Class for complex baseband signals.

    Baseband signals are Nyquist-sampled analytic signals. In the
    complex baseband representation, the signal is shifted to be
    centered around zero frequency and is a complex-valued signal. As a
    consequence, ``sample_rate == chan_bw``. All frequency channels are
    assumed to be upper side-band (that is, the signal isn't spectrally
    flipped).

    See the documentation for :py:class:`RadioSignal` for more details.

    Parameters
    ----------
    z : array-like
        The signal data. Must be at least 2-dimensional with shape
        `(nsample, nchan, ...)`.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : Quantity
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    freq_align : {'bottom', 'center', 'top'}, default: 'center'
        The alignment of channels relative to the ``center_freq``. Default
        is 'center' (as with odd-length complex DFTs). 'bottom' and
        'top' only have an effect when ``nchan`` is even.
    meta : dict, optional
        Any metadata that the user might want to attach to the signal.

    See Also
    --------
    Signal
    RadioSignal
    """

    _req_dtype = (np.complex128, np.complex64)

    def __init__(
        self,
        z,
        /,
        *,
        sample_rate,
        start_time=None,
        center_freq,
        freq_align="center",
        meta=None,
    ):

        super().__init__(
            z,
            sample_rate=sample_rate,
            start_time=start_time,
            center_freq=center_freq,
            chan_bw=sample_rate,
            freq_align=freq_align,
            meta=meta,
        )

    def to_intensity(self):
        """Converts baseband signal to intensities.

        Returns
        -------
        out : IntensitySignal
            The converted signal.
        """
        z = self.data.real ** 2 + self.data.imag ** 2
        return IntensitySignal.like(self, z)


class DualPolarizationSignal(BasebandSignal):
    """Class for dual-polarization complex baseband signals.

    See the documentation for :py:class:`RadioSignal` and
    :py:class:`BasebandSignal` for more details.

    Parameters
    ----------
    z : array-like
        The signal data. Must be at least 3-dimensional with shape
        ``(nsample, nchan, npol, ...)`` where ``npol = 2``.
    sample_rate : Quantity
        The number of samples per second. Must be in units of frequency.
    start_time : Time, optional
        The start time of the signal (that is, the time at the first
        sample of the signal). Default is None.
    center_freq : Quantity
        The frequency at the center of the signal's band. Must be in
        units of frequency.
    freq_align : {'bottom', 'center', 'top'}, default: 'center'
        The alignment of channels relative to the ``center_freq``. Default
        is 'center' (as with odd-length complex DFTs). 'bottom' and
        'top' only have an effect when ``nchan`` is even.
    pol_type : {'linear', 'circular'}
        The polarization type of the signal. 'linear' for linearly
        polarized signals (with basis [X, Y]) or 'circular' for
        circularly polarized signals (with basis [L, R]).
    meta : dict, optional
        Any metadata that the user might want to attach to the signal.

    See Also
    --------
    Signal
    RadioSignal
    BasebandSignal

    Notes
    -----
    Linear polarization is assumed to have basis ``[X, Y]`` and circular
    polarization is assumed to have basis ``[L, R]``. For example, with
    ``pol_type='circular'``, ``z[:, :, 0]`` refers to the right-handed
    circular polarization component.
    """

    _req_shape = (None, None, 2)
    _axes_labels = {"time": 0, "freq": 1, "pol": 2}

    def __init__(
        self,
        z,
        /,
        *,
        sample_rate,
        start_time=None,
        center_freq,
        freq_align="center",
        pol_type,
        meta=None,
    ):

        super().__init__(
            z,
            sample_rate=sample_rate,
            start_time=start_time,
            center_freq=center_freq,
            freq_align=freq_align,
            meta=meta,
        )

        self.pol_type = pol_type

    def _attr_repr(self):
        s = super()._attr_repr()
        pols = {"linear": "[X, Y]", "circular": "[L, R]"}
        pol_str = f"{self.pol_type} {pols[self.pol_type]}"
        s += f"Polarization Type: {pol_str}\n"
        return s

    @property
    def pol_type(self):
        """Polarization type (linear or circular)."""
        return self._pol_type

    @pol_type.setter
    def pol_type(self, pol_type):
        if pol_type in {"linear", "circular"}:
            self._pol_type = pol_type
        else:
            raise ValueError("pol_type must be in {'linear', 'circular'}")

    def to_linear(self):
        """Converts the dual-polarization signal to linear basis.

        If polarization basis is already linear, simply returns a copy of
        the signal object.

        Returns
        -------
        out : DualPolarizationSignal
            The converted signal.
        """
        if self.pol_type == "circular":
            axis = self.get_axis("pol")
            L = np.take(self.data, 0, axis=axis)
            R = np.take(self.data, 1, axis=axis)

            X = L + R
            Y = 1j * (L - R)

            z = np.stack([X, Y], axis=axis) / np.sqrt(2)
        else:
            z = self.data

        return type(self).like(self, z, pol_type="linear")

    def to_circular(self):
        """Converts the dual-polarization signal to circular basis.

        If polarization basis is already circular, simply returns a copy of
        the signal object.

        Returns
        -------
        out : DualPolarizationSignal
            The converted signal.
        """
        if self.pol_type == "linear":
            axis = self.get_axis("pol")
            X = np.take(self.data, 0, axis=axis)
            Y = np.take(self.data, 1, axis=axis)

            L = X - 1j * Y
            R = X + 1j * Y

            z = np.stack([L, R], axis=axis) / np.sqrt(2)
        else:
            z = self.data

        return type(self).like(self, z, pol_type="circular")

    def to_stokes(self):
        """Converts signal to IQUV Stokes representation.

        Returns
        -------
        out : FullStokesSignal
            Signal in Stokes IQUV representation.
        """
        axis = self.get_axis("pol")
        A = np.take(self.data, 0, axis=axis)
        B = np.take(self.data, 1, axis=axis)

        if self.pol_type == "linear":
            X, Y = A, B

            XX = X.real ** 2 + X.imag ** 2
            YY = Y.real ** 2 + Y.imag ** 2
            XY = X.conj() * Y

            i = XX + YY
            Q = XX - YY
            U = 2 * XY.real
            V = 2 * XY.imag

        elif self.pol_type == "circular":
            L, R = A, B

            LL = L.real ** 2 + L.imag ** 2
            RR = R.real ** 2 + R.imag ** 2
            LR = L.conj() * R

            i = LL + RR
            Q = 2 * LR.real
            U = 2 * LR.imag
            V = LL - RR

        z = np.stack([i, Q, U, V], axis=axis)
        return FullStokesSignal.like(self, z)



import os
import pulsarbat
import warnings

project = "pulsarbat"
copyright = "2022, Nikhil Mahajan"
author = "Nikhil Mahajan"
release = pulsarbat.__version__

json_url = "https://pulsarbat.readthedocs.io/en/dev/_static/switcher.json"
version_match = os.environ.get("READTHEDOCS_VERSION")
if not version_match or version_match.isdigit():
    json_url = "/_static/switcher.json"
    version_match = "dev"


# -- Extensions --------------------------------------------------------------

extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "matplotlib.sphinxext.plot_directive",
    "nb2plots",
    "texext",
    "numpydoc",
]


doctest_global_setup = plot_pre_code = """
import numpy as np
np.random.seed(123)
"""

# Plot
plot_include_source = True
plot_formats = [('png', 96)]
plot_html_show_formats = False
plot_html_show_source_link = False

# Autosummary
autosummary_generate = True

# Intersphinx mapping
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "baseband": ("https://baseband.readthedocs.io/en/latest/", None),
    "dask": ("https://docs.dask.org/en/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
}

# numpydoc
numpydoc_show_class_members = False
numpydoc_xref_param_type = True
numpydoc_xref_aliases = {
    # "array-like": ":term:`array-like <array_like>`",
    # "scalar": ":term:`scalar`",
    # "array": ":term:`array`",
    # "ndarray": "numpy.ndarray",
    "dtype": "numpy.dtype",
    "Quantity": "astropy.units.Quantity",
    "Time": "astropy.time.Time",
    "Unit": "astropy.units.Unit",
}


# -- General configuration ---------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_sourcelink_suffix = ""
html_theme_options = {
    "collapse_navigation": True,
    "navigation_depth": 2,
    "show_prev_next": False,
    "icon_links": [
        {
            "name": "Home Page",
            "url": "https://pulsarbat.readthedocs.io/",
            "icon": "fas fa-home",
            "type": "fontawesome",
        },
        {
            "name": "GitHub",
            "url": "https://github.com/theXYZT/pulsarbat",
            "icon": "fab fa-github-square",
            "type": "fontawesome",
        },
    ],
    "logo": {
        "image_light": "pulsarbat_light.svg",
        "image_dark": "pulsarbat_dark.svg",
    },
    "navbar_start": ["navbar-logo", "version-switcher"],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "page_sidebar_items": ["page-toc"],
    "switcher": {
        "json_url": json_url,
        "version_match": version_match,
    },
}

html_sidebars = {
    "**": ["search-field", "sidebar-nav-bs"],
}

html_context = {
    "default_mode": "light",
}

html_static_path = ["_static", ]


warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message="Matplotlib is currently using agg, which is a"
    " non-GUI backend, so cannot show the figure.",
)


def setup(app):
    app.add_css_file("custom.css")
    app.add_js_file("copybutton.js")



"""PULSAR Baseband Analysis Tools."""

# flake8: noqa

__author__ = """Nikhil Mahajan"""
__email__ = "mahajan@astro.utoronto.ca"
__version__ = "0.0.8"

from . import core
from .core import *

from . import transforms
from .transforms import *

from . import pulsar
from .pulsar import *

from . import fft
from . import utils
from . import readers
from . import contrib


__all__ = [
    "fft",
    "utils",
    "readers",
    "contrib",
]

__all__.extend(core.__all__)
__all__.extend(transforms.__all__)
__all__.extend(pulsar.__all__)



"""Tests for pulsar phase prediction."""

import pytest
from io import StringIO
from pathlib import Path
import numpy as np
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb
import cloudpickle

TEST_POLYCO = Path(__file__).parent.absolute() / "data" / "timing.dat"

ENTRY_TEXT = """B1937+21    7-May-18   0.00   58245.00000000000   71.020168
 146754136477.666475  641.928232294317   ao   90   12   327.000
 -3.17034199847385061e-07  2.76360291261698521e+00  8.05424212611731503e-05
 -1.14853014406135967e-07 -1.39769248548540950e-10  6.39552923641417649e-13
  3.19619782475082226e-15 -5.35166928586675360e-16 -4.58065943719761444e-19
  2.26855569952374124e-19 -4.63024751309515689e-23 -3.42478559749583972e-23
"""

INVALID = """B1937+21    7-May-18   93600.00   58245.40000000000   71.020167
 146776323107.982696  641.928232294317   ao  288   12   327.000
  2.53531626087091441e-08  2.71634485209303600e+00 -1.18579224383567124e-04
  2.13006534220737831e-08  1.82945349045995736e-10 -2.17093444265104196e-14
 -1.19143951096548092e-16  8.75742271090786040e-20  6.90038715651136990e-23
 -3.13982943744315582e-24 -4.22839364952768232e-28  5.14827898395789200e-29
B1937+21    7-May-18  163000.00   58245.68750000000            71.020167
 146792269715.610996  641.928232294317   ao   90   12   327.000
 -4.76040685223260378e-07  2.67040349305271674e+00  3.73410078415526979e-05
  1.57043805946084180e-07 -5.08960654800334953e-11 -3.36505250137997982e-13
 -1.89629091176250827e-14  1.96470551312769751e-16  9.40225741785451917e-18
 -9.15695855709336387e-20 -1.59591049042759423e-21  1.52097772879173001e-23
"""

phase_tols = dict(rtol=0, atol=1E-8 * u.cycle)


class TestPredictor:
    def test_basic(self):
        p = pb.PhasePredictor.from_polyco(TEST_POLYCO)
        assert np.all(pb.PhasePredictor(p) == p)
        assert len(p.intervals) == 1

        q = cloudpickle.loads(cloudpickle.dumps(p))
        assert np.all(q == p)

        t = Time("58245.375", format="mjd", precision=9)
        ph, f0, f1 = p(t), p.f0(t), p.f0(t, n=1)

        assert type(ph) is pb.Phase
        assert int(ph.int.cycle) == 146774936445
        assert np.isclose(ph.frac.cycle, 0.058161699852649296)
        assert u.isclose(f0, 641.973647812571 * u.cycle / u.s, rtol=1e-8)
        assert u.isclose(f1, -6.635997412662843e-08 * u.cycle / u.s ** 2, rtol=1e-8)

        ts = t + np.arange(10000) * u.us
        ph = p(ts)
        assert len(ph) == len(ts)
        assert int(ph[-1].int.cycle) == 146774936451
        assert np.isclose(ph[-1].frac.cycle, 0.4772562027766636)

        ts = Time("58244.91", format="mjd") + np.arange(1400) * u.min
        ph, f0 = p(ts), p.f0(ts)
        assert int(ph[-1].int.cycle) - int(ph[0].int.cycle) == 53887265
        assert np.all(f0 > 641.97269 * u.cycle / u.s)
        assert np.all(f0 < 641.97455 * u.cycle / u.s)
        assert f0.ptp() < (0.001858 * u.cycle / u.s)

        q = p[[0, 1, 2, 4, 5, 6, 8, 9]]
        assert len(q.intervals) == 3

        with pytest.raises(ValueError):
            t = Time("60000", format="mjd")
            _ = p(t)

    def test_stringio(self):
        p = pb.PhasePredictor.from_polyco(StringIO(ENTRY_TEXT))
        ref = pb.Phase(146754136477, 0.666475) + (-3.17034199847385061e-07)
        assert u.isclose(ref, p(p["tmid"][0]), **phase_tols)

        text = "this is not a polyco"
        with pytest.raises(ValueError):
            _ = pb.PhasePredictor.from_polyco(StringIO(text))

        with pytest.raises(ValueError):
            _ = pb.PhasePredictor.from_polyco(StringIO(INVALID))

    def test_phasepol(self):
        p = pb.PhasePredictor.from_polyco(StringIO(ENTRY_TEXT))
        t = p[0]["tmid"]
        pol, ref = p.phasepol(t)
        assert u.isclose(p(t + 1 * u.s), ref + pol(1), **phase_tols)
        assert u.isclose(p(t + 8 * u.s), ref + pol(8), **phase_tols)
        assert u.isclose(p(t + 1 * u.ms), ref + pol(0.001), **phase_tols)

        with pytest.raises(ValueError):
            _ = p.phasepol(t + np.arange(10) * u.ms)

    def test_polyco_entry(self):
        t = Time("60000", format="mjd")

        entry = pb.PolycoEntry(
            psr="Fake",
            obs="@",
            freq=1 * u.GHz,
            tmid=t,
            span=1 * u.day,
            rphase=0,
            poly=np.polynomial.Polynomial([0, 1]),
        )

        p = pb.PhasePredictor([entry])
        assert u.isclose(p(t), pb.Phase(0), **phase_tols)
        assert u.isclose(p(t + 1 * u.s), pb.Phase(1), **phase_tols)
        assert u.isclose(p(t + 1 * u.ms), pb.Phase(0.001), **phase_tols)
        assert u.isclose(p(t + 1 * u.hr), pb.Phase(3600), **phase_tols)

        assert u.isclose(p.f0(t), 1 * u.cycle / u.s)
        assert u.isclose(p.f0(t + 1 * u.hr), 1 * u.cycle / u.s)

        assert u.isclose(p.f0(t, n=1), 0 * u.cycle / u.s**2)
        assert u.isclose(p.f0(t + 1 * u.hr, n=1), 0 * u.cycle / u.s**2)


class TestPhase:
    def test_basic(self):
        a = pb.Phase(1.2)
        repr(a)
        str(a)
        assert a.isscalar
        assert u.allclose(a.int, 1 * u.cycle)
        assert u.allclose(a.frac, 0.2 * u.cycle)
        assert u.allclose(pb.FractionalPhase(a), 0.2 * u.cycle)
        assert u.allclose(a.cycle, 1.2 * u.cycle)
        assert u.allclose(a.to(u.deg), 432 * u.deg)

        a = pb.Phase([1.2, 1.8, 2.6])
        repr(a)
        str(a)
        assert u.allclose(a.int, [1, 2, 3] * u.cycle)
        assert u.allclose(a.frac, [0.2, -0.2, -0.4] * u.cycle)
        assert u.allclose(pb.FractionalPhase(a), [0.2, -0.2, -0.4] * u.cycle)
        assert u.allclose(a.cycle, [1.2, 1.8, 2.6] * u.cycle)
        assert u.allclose(a.to(u.deg), [432, 648, 936] * u.deg)

        assert np.array_equal(np.int64(a.int.value), np.array([1, 2, 3]))

        a = pb.Phase(np.arange(10))
        assert u.allclose(a.astype(np.int64), np.arange(10) * u.cycle)
        assert u.allclose(a.astype(np.float64), np.arange(10) * u.cycle)

    def test_from_string(self):
        def check_phase(ph, i, f):
            assert np.int64(ph.int.value) == i
            assert np.float64(ph.frac.value) == f

        assert u.allclose(pb.Phase(["49.64", "12.34"]), pb.Phase([49.64, 12.34]))

        s = "23424249845.323429437454"
        sx = [s, f"{s}E2", f"{s}e2", f"{s}E+2", f"{s}E-2"]
        ix = np.int64(
            [23424249845, 2342424984532, 2342424984532, 2342424984532, 234242498]
        )
        fx = np.float64(
            [0.323429437454, 0.3429437454, 0.3429437454, 0.3429437454, 0.45323429437454]
        )

        for s, i, f in zip(sx, ix, fx):
            check_phase(pb.Phase.from_string(s), i, f)

        ph = pb.Phase.from_string(sx)
        assert np.array_equal(np.int64(ph.int.value), ix)
        assert np.array_equal(np.float64(ph.frac.value), fx)

    def test_math_operations(self):
        a = pb.Phase([1.2, 1.8, 2.6]) + pb.Phase([3.4, 5.6, 2.3])
        b = pb.Phase([4.6, 7.4, 4.9])
        assert u.allclose(a, b)

        a = pb.Phase(1.2)
        assert u.isclose(4 * a, pb.Phase(4 * 1.2))
        assert u.isclose(a + 0.8, pb.Phase(1.2 + 0.8))

        b = pb.Phase(0.3, 0.5)
        assert u.isclose(a + b, pb.Phase(2.0))

        b = pb.Phase(pb.Phase(0.3), pb.Phase(0.5))
        assert u.isclose(a + b, pb.Phase(2.0))

        assert u.isclose(pb.Phase(100 / 3) + pb.Phase(10 / 3), pb.Phase(110 / 3))
        assert u.isclose(np.exp(1j * pb.Phase(0.5)), -1)
        assert u.isclose(np.exp(1j * pb.Phase(1.0)), +1)

    def test_common_operations(self):
        ph = pb.Phase([1.3, 2.3, 12.7, 0.1, 5.6, 6.8])
        assert ph.min() == pb.Phase(0.1)
        assert ph.max() == pb.Phase(12.7)
        ph2 = ph.sort()
        assert ph2[0] == pb.Phase(0.1)
        assert ph2[-1] == pb.Phase(12.7)



"""Core signal transforms."""

import numpy as np
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb
import functools
import dask.array as da

__all__ = [
    "signal_transform",
    "concatenate",
    "time_shift",
    "fast_len",
]


def signal_transform(func):
    """Wraps an array function and returns a signal transform.

    The function being decorated must accept an array and return an array
    with function signature::

        func(x, *args, /, **kwargs)

    where `x` is the array being transformed. The decorated function will
    accept a Signal object in place of `x` and by default return a Signal object
    of the same type (unless a different `signal_type` is specified). The
    properties of the returned signal can be modified via `signal_kwargs` if needed.

    If the Signal data is contained within a Dask array, `func` will be applied
    to every chunk independently using `dask.array.map_blocks`. Keyword arguments
    specific to `map_blocks` can be passed via `dask_kwargs`.
    """
    @functools.wraps(func)
    def wrapper(x, *args, signal_type=None, signal_kwargs=dict(),
                dask_kwargs=dict(), **kwargs):

        sig_class = type(x) if signal_type is None else signal_type
        if not issubclass(sig_class, pb.Signal):
            raise TypeError("Signal type must be a subclass of pulsarbat.Signal!")

        if isinstance(x.data, da.Array):
            z = da.map_blocks(func, x.data, **dask_kwargs, **kwargs)
        else:
            z = func(x.data, **kwargs)

        return sig_class.like(x, z, **signal_kwargs)

    return wrapper


def concatenate(signals, /, axis=0):
    """Concatenates multiple signals along given axis.

    Signals must be contiguous along the axis of concatenation. The
    concatenated signal will inherit attributes from given `kwargs` and
    then from the first signal in the sequence, `signals[0]`, except for
    `center_freq` and `freq_align` when concatenating along frequency
    (which must be computed accordingly).

    Parameters
    ----------
    signals : sequence of `~Signal`-like
        Sequence of signals to concatenate. All signals must be Signal
        objects and have the same type and `sample_rate`. If concatenating
        along frequency, then they must also have the same `chan_bw`.
    axis : int or 'time' or 'freq', optional
        Axis along which to concatenate signals. Default is `0`. `time`
        is an alias for `0` (concatenating along time). `freq` implies
        `axis=1` (concatenating along frequency) and requires that
        signals are instances of RadioSignal.

    Returns
    -------
    out : `~Signal`-like
        Concatenated signal of same type as input signals.
    """
    try:
        sig_type = type(signals[0])
    except IndexError:
        raise ValueError("Need at least one signal to concatenate.")
    else:
        if not issubclass(sig_type, pb.Signal):
            raise TypeError("Signals must be pulsarbat.Signal objects.")

        if not all(type(s) == sig_type for s in signals):
            raise TypeError("All signals must have same type!")

    ref_sr = signals[0].sample_rate
    if not all(u.isclose(ref_sr, s.sample_rate) for s in signals):
        raise ValueError("Signals must have the same sample_rate!")

    ref_st = None
    if axis in {0, "time"}:
        n = 0
        for s in signals:
            if s.start_time is not None:
                if ref_st is None:
                    ref_st = s.start_time - (n / ref_sr)
                elif not Time.isclose(ref_st + (n / ref_sr), s.start_time):
                    raise ValueError("Signals not contiguous in time.")
            n += len(s)
        axis = 0
    else:
        for s in signals:
            if s.start_time is not None:
                if ref_st is None:
                    ref_st = s.start_time
                elif not Time.isclose(ref_st, s.start_time):
                    raise ValueError("Signals have different start_time.")

    kw = {"start_time": ref_st}

    if isinstance(signals[0], pb.RadioSignal):
        ref_cbw = signals[0].chan_bw
        if not all(u.isclose(ref_cbw, s.chan_bw) for s in signals):
            raise ValueError("RadioSignals must have the same chan_bw!")

        if axis in {1, "freq"}:
            for x, y in zip(signals, signals[1:]):
                chan_diff = y.channel_freqs[0] - x.channel_freqs[-1]
                if not u.isclose(chan_diff, ref_cbw):
                    raise ValueError("Signals not contiguous in frequency.")

            f0, f1 = signals[0].channel_freqs[0], signals[-1].channel_freqs[-1]
            axis = 1
        else:
            ref_cfs = signals[0].channel_freqs
            if not all(u.allclose(ref_cfs, s.channel_freqs) for s in signals):
                raise ValueError("Signals have different frequency channels.")
            f0, f1 = ref_cfs[0], ref_cfs[-1]

        kw["center_freq"] = (f0 + f1) / 2
        kw["freq_align"] = "center"

    elif axis == "freq":
        err = "Signals must be pb.RadioSignal objects when axis is 'freq'."
        raise TypeError(err)

    z = np.concatenate([s.data for s in signals], axis=axis)
    return sig_type.like(signals[0], z, **kw)


def time_shift(z, /, t):
    """Shift signal by given number of samples or time.

    This function shifts signals in time via FFT by multiplying by
    a phase gradient in frequency domain.

    Parameters
    ----------
    z : `~Signal`
        Input signal.
    t : int, float or `~astropy.units.Quantity`
        Shift amount. If a number (int or float), the signal is shifted
        by that number of samples. An astropy Quantity with units of
        time can also be passed, in which case the signal will be
        shifted by `t * z.sample_rate` samples.

    Returns
    -------
    out : `~Signal`
        Shifted signal.
    """
    if t == 0:
        return z

    if isinstance(t, u.Quantity):
        n = np.float64((t * z.sample_rate).to_value(u.one))
    else:
        n = np.float64(t)

    if isinstance(z.data, da.Array):
        f = da.fft.fftfreq(len(z), 1)
    else:
        f = np.fft.fftfreq(len(z), 1)

    ix = tuple(slice(None) if i == 0 else None for i in range(z.ndim))
    ph = np.exp(-2j * np.pi * n * f).astype(np.complex64)[ix]
    shifted = pb.fft.ifft(pb.fft.fft(z.data, axis=0) * ph, axis=0)

    if np.iscomplexobj(z.data):
        shifted = shifted.astype(z.dtype)
    else:
        shifted = shifted.real.astype(z.dtype)

    x = type(z).like(z, shifted, start_time=z.start_time - n * z.dt)

    if n >= 0:
        i = np.int64(np.ceil(n))
        x = x[i:]
    else:
        i = np.int64(np.floor(n))
        x = x[:i]

    return x


def fast_len(z, /):
    """Crops signal to an efficient length for FFTs.

    Output signal is cropped to a length of the largest 7-smooth number
    less than or equal to the length of the input signal.

    Parameters
    ----------
    z : `~Signal`
        Input signal.

    Returns
    -------
    out : `~Signal`
        Cropped signal.
    """
    N = len(z)
    fast_N = pb.utils.prev_fast_len(N)
    return z[:fast_N]



"""Transforms: Functions that process Signals into other Signals."""

# flake8: noqa

from . import transforms
from .transforms import *

from . import dedispersion
from .dedispersion import *

__all__ = transforms.__all__.copy()
__all__ += dedispersion.__all__



"""Tests for `pulsarbat.utils`"""

import pytest
import bisect
import numpy as np
from pulsarbat.utils import real_to_complex, next_fast_len, prev_fast_len

FAST = [
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 14, 15, 16, 18, 20, 21, 24, 25, 27, 28, 30, 32,
    35, 36, 40, 42, 45, 48, 49, 50, 54, 56, 60, 63, 64, 70, 72, 75, 80, 81, 84, 90, 96,
    98, 100, 105, 108, 112, 120, 125, 126, 128, 135, 140, 144, 147, 150, 160, 162, 168,
    175, 180, 189, 192, 196, 200, 210, 216, 224, 225, 240, 243, 245, 250, 252, 256, 270,
    280, 288, 294, 300, 315, 320, 324, 336, 343, 350, 360, 375, 378, 384, 392, 400, 405,
    420, 432, 441, 448, 450, 480, 486, 490, 500, 504, 512, 525, 540, 560, 567, 576, 588,
    600, 625, 630, 640, 648, 672, 675, 686, 700, 720, 729, 735, 750, 756, 768, 784, 800,
    810, 840, 864, 875, 882, 896, 900, 945, 960, 972, 980, 1000,
]


class TestRealToComplex:
    def gen_input(self, t, w, p):
        return np.cos(w * t + p)

    def prediction(self, t, w, p):
        return np.exp(1j * ((w - (len(t) / 4)) * t[::2] + p))

    def test_theoretical(self):
        for N in [511, 512]:
            t = np.linspace(0, 2 * np.pi, N, endpoint=False)
            for w in [1, 2, 127, 128, 129, 254, 255]:
                for p in [-np.pi, -np.pi / 2, 0, np.pi / 2]:
                    x = self.gen_input(t, w, p)
                    y = self.prediction(t, w, p)
                    z = real_to_complex(x)
                    assert np.allclose(y, z)

    def test_empty(self):
        for sample_shape in [(), (2,), (4, 4)]:
            x = np.zeros((0,) + sample_shape)
            y = real_to_complex(x, axis=0)
            assert np.array_equal(x, y)

    def test_axis(self):
        N = 128
        t = np.linspace(0, 2 * np.pi, N, endpoint=False)
        ws = [1, 2, 3, 4]

        x = np.stack([self.gen_input(t, w, 0) for w in ws], axis=0)
        y = np.stack([self.prediction(t, w, 0) for w in ws], axis=0)
        z = real_to_complex(x, axis=1)
        assert np.allclose(y, z)

        x = np.stack([self.gen_input(t, w, 0) for w in ws], axis=1)
        y = np.stack([self.prediction(t, w, 0) for w in ws], axis=1)
        z = real_to_complex(x, axis=0)
        assert np.allclose(y, z)

    def test_bad_args(self):
        x = np.ones((128, 4), dtype=complex)
        with pytest.raises(ValueError):
            _ = real_to_complex(x, axis=0)

    def test_dtype(self):
        d = [(np.float32, np.complex64), (np.float64, np.complex128)]
        for in_type, out_type in d:
            x = np.random.default_rng().standard_normal(32)
            y = real_to_complex(x.astype(in_type))
            assert y.dtype == out_type


class TestFastLengths:
    def test_next_fast_len(self):
        assert next_fast_len(0) == 0

        for i in range(1, 1000):
            assert next_fast_len(i) == FAST[bisect.bisect_left(FAST, i)]

    def test_prev_fast_len(self):
        assert prev_fast_len(0) == 0

        for i in range(1, 1000):
            assert prev_fast_len(i) == FAST[bisect.bisect(FAST, i) - 1]



"""Baseband reader classes."""

import numpy as np
import astropy.units as u
from astropy.time import Time
from contextlib import nullcontext
import baseband
import pulsarbat as pb
from pulsarbat.readers import BaseReader

__all__ = [
    "BasebandReader",
    "GUPPIRawReader",
    "DADAStokesReader",
]


class BasebandReader(BaseReader):
    """Wrapper around StreamReader from the `~baseband` package.

    Parameters
    ----------
    name, **kwargs
        Arguments to pass on to `~baseband.open` to create a StreamReader
        object via `baseband.open(name, 'rs', **kwargs)`.
    signal_type : class, optional
        Type of signal that will be returned by `read()`. Default is
        `Signal`. Accepted values are subclasses of `Signal`.
    signal_kwargs : dict, optional
        Additional `kwargs` to pass on to `signal_type` when creating a
        Signal object. Must not include `sample_rate` or `start_time` as
        dictionary fields.
    intensity : bool, optional
        Whether the data is intensity data. If `signal_type` is a
        subclass of `pb.IntensitySignal`, assumed to be True. If
        `signal_type` is a subclass of `pb.BasebandSignal`, assumed to be
        False. Default is False.
    lower_sideband : bool or array-like, optional
        Whether the data is lower-sideband (LSB) data. Default is False.
        If not a boolean, must be an array-like of booleans with the
        same shape as `sample_shape` of original data as read by
        `StreamReader` in the `baseband` package.
    """

    def __init__(
        self,
        name,
        /,
        *,
        signal_type=pb.Signal,
        signal_kwargs=dict(),
        intensity=None,
        lower_sideband=False,
        **kwargs,
    ):

        self._name = name
        self._kwargs = kwargs

        if intensity is None:
            self._intensity = issubclass(signal_type, pb.IntensitySignal)
        else:
            self._intensity = bool(intensity)

            if issubclass(signal_type, pb.BasebandSignal) and self.intensity:
                raise ValueError("intensity must be False when using pb.BasebandSignal")

            if issubclass(signal_type, pb.IntensitySignal) and not self.intensity:
                raise ValueError("intensity must be True when using pb.IntensitySignal")

        with self._get_fh() as fh:
            self._complex_data = bool(fh.complex_data)
            self._in_sample_shape = fh.shape[1:]

            if self.intensity and self.complex_data:
                raise ValueError("Intensity data cannot be complex-valued!")

            if self.real_baseband:
                _sr = (fh.sample_rate / 2).to(u.MHz)
                _length = fh.shape[0] // 2
                _dtype = np.complex64
            else:
                _sr = fh.sample_rate.to(u.MHz)
                _length = fh.shape[0]
                _dtype = np.complex64 if self.complex_data else np.float32

            _t0 = Time(fh.start_time, format="isot", precision=9)

        self.lower_sideband = lower_sideband
        self._dtype = np.dtype(_dtype)

        # Determine sample shape by reading a dummy array
        _shape = (_length,) + self._read_array(0, 0).shape[1:]

        super().__init__(
            shape=_shape,
            dtype=_dtype,
            signal_type=signal_type,
            sample_rate=_sr,
            start_time=_t0,
            **signal_kwargs,
        )

    def _get_fh(self):
        return baseband.open(self._name, "rs", **self._kwargs)

    @property
    def complex_data(self):
        """Whether the data is complex-valued."""
        return self._complex_data

    @property
    def intensity(self):
        """Where the data is intensity data (as opposed to raw baseband)."""
        return self._intensity

    @property
    def real_baseband(self):
        """Whether the data is real-valued baseband data."""
        return not (self.intensity or self.complex_data)

    @property
    def lower_sideband(self):
        """Whether data is lower sideband (LSB) data."""
        return self._lower_sideband

    @lower_sideband.setter
    def lower_sideband(self, s):
        if type(s) != bool:
            s = np.array(s).astype(bool)
            if s.shape != self._in_sample_shape:
                err = f"Got {s.shape}, expected {self._in_sample_shape}"
                raise ValueError(f"Invalid lower_sideband shape. {err}")

        self._lower_sideband = s

    def _read_baseband(self, offset, n, /, lock=nullcontext(), **kwargs):
        """Read n samples from given offset using baseband."""
        with lock:
            with self._get_fh() as fh:
                if self.real_baseband:
                    fh.seek(2 * offset)
                    z = pb.utils.real_to_complex(fh.read(2 * n), axis=0)
                else:
                    fh.seek(offset)
                    z = fh.read(n)

        if not self.intensity:
            if self.lower_sideband is True:
                z = z.conj()
            elif self.lower_sideband is not False:
                z[:, self.lower_sideband] = z[:, self.lower_sideband].conj()

        return z.astype(self.dtype, copy=False)

    def _read_array(self, offset, n, /, **kwargs):
        """Read n samples from given offset into numpy array.

        Post-processing for specific formats (such as correcting the
        order of data dimensions) should be done in this method by
        subclasses.
        """
        return self._read_baseband(offset, n, **kwargs)

    def read(self, offset, n, /, **kwargs):
        """Read `n` samples from given offset.

        Parameters
        ----------
        offset : int
            Position to read from. Must be non-negative.
        n : int
            Number of samples to read. Must be non-negative.
        **kwargs
            Additional keyword arguments. Currently supported are:
              * `use_dask` -- Whether to use dask arrays.
              * `lock` -- A lock object to prevent concurrent reads.
                          Must be a context manager.

        Returns
        -------
        out : `~Signal` object or subclass
            Signal of length `n` containing data that was read.
        """
        return super().read(offset, n, **kwargs)


class GUPPIRawReader(BasebandReader):
    """Baseband reader for GUPPI raw voltage data format.

    Parameters
    ----------
    name
        Filename, filehandle, or sequence of filenames to pass on to
        `~baseband.open` to create a GUPPIStreamReader object via
        ``baseband.open(name, 'rs', format='guppi', squeeze=False)``.
    """

    def __init__(self, name, /):
        kwargs = {"format": "guppi", "squeeze": False}

        with baseband.open(name, "rs", **kwargs) as fh:
            header = fh.header0
            obsfreq = header["OBSFREQ"] * u.MHz
            pol = {"LIN": "linear", "CIRC": "circular"}[header["FD_POLN"]]

        signal_kwargs = {
            "center_freq": obsfreq,
            "freq_align": "center",
            "pol_type": pol,
        }

        super().__init__(
            name,
            signal_type=pb.DualPolarizationSignal,
            signal_kwargs=signal_kwargs,
            lower_sideband=not header.sideband,
            **kwargs,
        )

    def _read_array(self, offset, n, /, **kwargs):
        """Read n samples from current read position into numpy array."""
        z = self._read_baseband(offset, n, **kwargs)
        return z.transpose(0, 2, 1)


class DADAStokesReader(BasebandReader):
    """Reader for full Stokes data in DADA format.

    Parameters
    ----------
    name
        File name, filehandle, or sequence of file names to pass on to
        `~baseband.open` to create a DADAStreamReader object via
        `baseband.open(name, 'rs', format='dada')`.
    """

    def __init__(self, name, /):
        kwargs = {"format": "dada", "squeeze": False}

        with baseband.open(name, "rs", **kwargs) as fh:
            header = fh.header0

            if not (header["NPOL"] == 4 and header["NDIM"] == 1):
                raise ValueError("Does not look like Full Stokes data")

            lsb = header["BW"] < 0
            freq = header["FREQ"] * u.MHz
            chan_bw = abs(header["BW"] / header["NCHAN"]) * u.MHz
            freq_align = "top" if lsb else "bottom"

        signal_kwargs = {
            "center_freq": freq,
            "chan_bw": chan_bw,
            "freq_align": freq_align,
        }

        super().__init__(
            name,
            signal_type=pb.FullStokesSignal,
            signal_kwargs=signal_kwargs,
            lower_sideband=lsb,
            **kwargs,
        )

    def _read_array(self, offset, n, /, **kwargs):
        """Read n samples from current read position into numpy array."""
        z = self._read_baseband(offset, n, **kwargs)

        if self.lower_sideband:
            z = np.flip(z, axis=-1)

        return z.transpose(0, 2, 1)



"""Tests for dedispersion."""

import pytest
import numpy as np
import astropy.units as u
import pulsarbat as pb
from astropy.time import Time
import scipy.signal
import dask.array as da


class TestDispersionMeasure:
    def test_basic(self):
        DM = pb.DispersionMeasure(2.41e-4)

        for f in [0.1, 1.0, 10.0]:
            dt = DM.time_delay(f * u.MHz, np.inf)
            assert u.isclose(dt, (1 / f / f) * u.s)

            dt = DM.time_delay(np.inf, f * u.MHz)
            assert u.isclose(dt, -(1 / f / f) * u.s)

        dt = DM.time_delay(2 * u.MHz, 1 * u.MHz)
        assert u.isclose(dt, -0.75 * u.s)

        for SR in [1 * u.MHz, 10 * u.MHz, 1 * u.kHz]:
            dn = DM.sample_delay(1 * u.MHz, np.inf, SR)
            assert np.isclose(dn, (SR * u.s).to_value(u.one))

        for a in [10, 20, 100]:
            DM = pb.DispersionMeasure(2.41e-4 * a)
            assert u.isclose(DM.time_delay(1 * u.MHz, np.inf), a * u.s)


class TestCoherentDedispersion:
    @pytest.mark.parametrize("dm", [10.0, 50.0, 100.0])
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_basic(self, dm, use_dask):
        shape = (8192, 4)
        fcen = 1 * u.GHz
        SR = 1 * u.MHz
        DM = pb.DispersionMeasure(dm)

        if use_dask:
            f = da.random.standard_normal
        else:
            f = np.random.standard_normal

        x = f(shape) + 1j * f(shape)
        with pytest.raises(TypeError):
            z = pb.Signal(x, sample_rate=SR)
            _ = pb.coherent_dedispersion(z, DM)

        with pytest.raises(TypeError):
            z = pb.Signal(x, sample_rate=SR)
            _ = DM.chirp_from_signal(z)

        z = pb.BasebandSignal(
            x, sample_rate=SR, center_freq=fcen, start_time=Time.now()
        )

        for ref_freq in [z.min_freq, z.center_freq, z.max_freq]:
            y = pb.coherent_dedispersion(z, DM, ref_freq=ref_freq)

            assert len(z) - len(y) >= DM.sample_delay(
                z.min_freq, z.max_freq, z.sample_rate
            )

            d_st = y.start_time - z.start_time
            toffset = int(np.rint((d_st * z.sample_rate).to_value(u.one)))
            assert toffset >= DM.sample_delay(ref_freq, z.max_freq, z.sample_rate)

    @pytest.mark.parametrize("seed", [4, 8, 15, 16, 23, 42])
    def test_reversibility(self, seed):
        ref_freq = 600 * u.MHz
        SR = 400 * u.MHz
        t0 = Time(56000.0, format="mjd")
        DM = pb.DispersionMeasure(0.01)

        N, M = 2 ** 18, 2 ** 12
        R = np.random.default_rng(seed=seed)
        x = R.standard_normal(N) + 1j * R.standard_normal(N)
        x *= np.exp(-(((np.arange(N) - N // 2) / M) ** 2))

        sos = scipy.signal.butter(10, 0.45, "lowpass", fs=1.0, output="sos")
        x = scipy.signal.sosfilt(sos, x)

        sig = pb.BasebandSignal(
            x.reshape(-1, 1), sample_rate=SR, center_freq=ref_freq, start_time=t0
        )
        temp = pb.coherent_dedispersion(sig, DM)
        sig2 = pb.coherent_dedispersion(temp, -DM)

        toffset = sig2.start_time - sig.start_time
        noffset = int(np.rint((toffset * sig.sample_rate).to(u.one)))
        sig1 = sig[noffset : noffset + len(sig2)]
        res = np.array(sig1) - np.array(sig2)
        assert np.allclose(res, 0, atol=3e-8)

    @pytest.mark.parametrize("DM", [pb.DM(0.01), pb.DM(0.02)])
    def test_correctness(self, DM):
        def gabor_wavelet(t, t0, f0, ts):
            a = 2j * np.pi * (t - t0) * f0 - ((t - t0) / ts) ** 2
            return np.exp(a.to_value(u.one))

        ref_freq = 600 * u.MHz
        SR = 400 * u.MHz

        index = 100000
        N = 2 ** 18
        t = np.arange(N) / SR
        t0 = t[index]

        width = 256
        x = np.zeros(N, dtype=np.complex128)
        for df in np.linspace(-3 * SR / 8, 3 * SR / 8, 13):
            dt = DM.time_delay(ref_freq + df, ref_freq)
            x += gabor_wavelet(t, t0 + dt, df, width / SR)

        sig1 = pb.BasebandSignal(
            x.reshape(-1, 1),
            sample_rate=SR,
            center_freq=ref_freq,
            start_time=Time.now(),
        )
        sig2 = pb.coherent_dedispersion(sig1, DM)

        toffset = sig2.start_time - sig1.start_time
        noffset = int(np.rint((toffset * sig2.sample_rate).to(u.one)))

        sig1, sig2 = np.array(sig1), np.array(sig2)
        id1, id2 = index - noffset - 8 * width, index - noffset + 8 * width

        sig1_power = (np.abs(sig1) ** 2).sum()
        sig2_power = (np.abs(sig2[id1:id2]) ** 2).sum()

        assert np.allclose(sig1_power, sig2_power)
        assert np.allclose(sig2[id2:], 0)
        assert np.allclose(sig2[:id1], 0)

    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize("DM", [pb.DM(10), pb.DM(20), pb.DM(50)])
    def test_chirp(self, use_dask, DM):
        shape = (8192, 4)

        if use_dask:
            f = da.random.standard_normal
        else:
            f = np.random.standard_normal

        x = f(shape) + 1j * f(shape)
        z = pb.BasebandSignal(x, sample_rate=1 * u.MHz, center_freq=1 * u.GHz)

        chirp = DM.chirp_from_signal(z)

        y1 = pb.coherent_dedispersion(z, DM)
        y2 = pb.coherent_dedispersion(z, DM, chirp=chirp)

        for rf in [z.min_freq, z.max_freq]:
            chirp = DM.chirp_from_signal(z, ref_freq=rf)
            y1 = pb.coherent_dedispersion(z, DM, ref_freq=rf)
            y2 = pb.coherent_dedispersion(z, DM, ref_freq=rf, chirp=chirp)
            assert np.allclose(y1, y2)


class TestIncoherentDedispersion:
    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize("dm", [50.0, 100.0, 200.0])
    def test_basic(self, dm, use_dask):
        SR = 1 * u.kHz
        ref_freq = 1 * u.GHz
        shape = (8192, 32, 4)
        t0 = Time.now()

        DM = pb.DispersionMeasure(dm)
        if use_dask:
            x = da.random.standard_normal(shape)
        else:
            x = np.random.standard_normal(shape)

        z1 = pb.FullStokesSignal(
            x, sample_rate=SR, start_time=t0, center_freq=ref_freq, chan_bw=8 * u.MHz
        )
        z2 = pb.incoherent_dedispersion(z1, DM)

        delay_top = DM.sample_delay(ref_freq, z1.channel_freqs[-1], SR).round()
        delay_bot = DM.sample_delay(z1.channel_freqs[0], ref_freq, SR).round()
        assert len(z1) - len(z2) == int(delay_top + delay_bot)



"""Tests for readers in `pulsarbat.reader`."""

import pytest
import numpy as np
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb
import pulsarbat.readers as pbr


class IndexReader(pbr.BaseReader):
    """A reader that returns indices of samples."""

    def __init__(
        self,
        /,
        *,
        shape,
        dtype=np.int32,
        sample_rate=1 * u.Hz,
        start_time=None,
        signal_type=pb.Signal,
        **signal_kwargs,
    ):
        super().__init__(
            signal_type=signal_type,
            shape=shape,
            dtype=dtype,
            sample_rate=sample_rate,
            start_time=start_time,
            **signal_kwargs,
        )

    def _read_array(self, offset, n, /):
        x = np.arange(offset, offset + n)
        x = x.reshape((-1,) + (self.ndim - 1) * (1,))
        x = x * np.ones(self.sample_shape)
        return x.astype(self.dtype)


class TestBaseReader:
    def test_basic_functionality(self):
        shape = (100, 4)
        dtype = np.uint16
        t0 = Time.now()
        SR = 1 * u.Hz

        r = IndexReader(shape=shape, dtype=dtype, sample_rate=SR, start_time=t0)

        _ = repr(r)
        _ = str(r)
        _ = dir(r)

        assert r.shape == shape
        assert len(r) == shape[0]
        assert r.ndim == len(shape)
        assert r.sample_shape == shape[1:]
        assert r.dtype == dtype

        assert u.isclose(SR, r.sample_rate)
        assert u.isclose(1 / SR, r.dt)
        assert u.isclose(shape[0] / SR, r.time_length)

        assert Time.isclose(t0, r.start_time)
        assert Time.isclose(t0 + shape[0] / SR, r.stop_time)

        assert r.offset_at(t0) == 0
        assert r.offset_at(r.stop_time) == len(r)

        t = r.time_at(0, unit=u.s)
        assert t.unit == u.s
        assert u.isclose(t, 0 * u.s)

        t = r.time_at(60, unit=u.min)
        assert t.unit == u.min
        assert u.isclose(t, 1 * u.min)

        for offset, n in [(0, 1), (4, 10), (10, 49)]:
            assert r.offset_at(offset / SR) == offset

            t = r.time_at(offset)
            assert Time.isclose(t, t0 + offset / SR)

            x = r.read(offset, n)
            assert type(x) == pb.Signal
            assert x.dtype == dtype
            assert x.sample_shape == r.sample_shape
            assert np.allclose(np.array(x), np.arange(offset, offset + n)[:, None])

        for temp in [r.start_time - 10 * u.s, r.stop_time + 10 * u.s]:
            with pytest.raises(EOFError):
                _ = r.offset_at(temp)

        with pytest.raises(EOFError):
            _ = r.read(99, 2)

        with pytest.raises(ValueError):
            _ = r.read(-1, 10)

        with pytest.raises(ValueError):
            _ = r.read(10, -1)

    def test_contains(self):
        shape = (4,)
        t0 = Time("2020-01-01T12:34:56.000", format="isot", precision=9)
        ts = t0 + np.arange(-2, 4 + 2) * u.s

        r = IndexReader(
            shape=shape, dtype=np.uint16, sample_rate=1 * u.Hz, start_time=t0
        )

        assert t0 + 2 * u.s in r
        assert all(r.contains(ts) == [0, 0, 1, 1, 1, 1, 0, 0])

        r = IndexReader(shape=shape, dtype=np.uint16, sample_rate=1 * u.Hz)
        assert not (t0 + 2 * u.s in r)
        assert all(r.contains(ts) == [0] * 8)

    @pytest.mark.parametrize(
        "sigtype, dtype, shape, sigkw",
        [
            (
                pb.DualPolarizationSignal,
                np.complex64,
                (1024, 4, 2),
                {"center_freq": 1 * u.GHz, "pol_type": "linear"},
            ),
            (
                pb.FullStokesSignal,
                np.float64,
                (1024, 4, 4),
                {"center_freq": 1 * u.GHz, "chan_bw": 1 * u.MHz},
            ),
        ],
    )
    def test_extra_functionality(self, sigtype, dtype, shape, sigkw):
        SR = 1 * u.MHz
        r = IndexReader(
            shape=shape,
            dtype=dtype,
            sample_rate=SR,
            start_time=None,
            signal_type=sigtype,
            **sigkw,
        )

        assert r.start_time is None
        assert r.stop_time is None

        offset, n = 5, 17
        x = r.read(offset, n)

        assert type(x) == sigtype
        assert u.allclose(x.center_freq, sigkw["center_freq"])
        assert x.start_time is None

        y = r.dask_read(offset, n)
        assert np.allclose(np.array(x), np.array(y))
        assert u.allclose(x.channel_freqs, y.channel_freqs)

    def test_bad_arguments(self):
        with pytest.raises(ValueError):
            _ = IndexReader(shape=())

        with pytest.raises(ValueError):
            _ = IndexReader(shape=(100,), signal_type=np.ndarray)

        for SR in (-5 * u.MHz, "fish", [10, 20] * u.Hz):
            with pytest.raises(ValueError):
                _ = IndexReader(shape=(100,), sample_rate=SR)

        for t0 in (1 * u.s, "fish", Time([424.23, 23424.42], format="unix")):
            with pytest.raises(ValueError):
                _ = IndexReader(shape=(100,), start_time=t0)

    def test_broken_reader(self):
        class BrokenReader(pbr.BaseReader):
            """A reader that returns indices of samples."""

            def __init__(self, /, *, shape=(100,), dtype=np.int32, N=0):
                self._N = N
                super().__init__(shape=shape, dtype=dtype, sample_rate=1 * u.Hz)

            def _read_array(self, offset, n, /):
                return np.zeros(self._N, dtype=np.int32)

            def read(self, offset, n, /, **kwargs):
                return super().read(offset, n, **kwargs)

        with pytest.raises(ValueError):
            _ = BrokenReader(dtype=np.float64)

        with pytest.raises(ValueError):
            _ = BrokenReader(shape=(100, 4, 2))

        with pytest.raises(ValueError):
            _ = BrokenReader(N=10)



"""Tests for core signal functions."""

import pytest
import numpy as np
import dask.array as da
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb


def assert_equal_signals(x, y):
    assert np.allclose(np.array(x), np.array(y)), f"{x.shape}, {y.shape}"
    assert Time.isclose(x.start_time, y.start_time)
    assert u.isclose(x.sample_rate, y.sample_rate)


def assert_equal_radiosignals(x, y):
    assert_equal_signals(x, y)
    assert u.allclose(x.channel_freqs, y.channel_freqs)


class TestSTFT:
    @pytest.mark.parametrize("shape", [(4224, 4, 2), (4233, 3, 2)])
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_reversibility(self, use_dask, shape):
        p = da if use_dask else np
        kw = {
            "sample_rate": 4 * u.MHz,
            "center_freq": 400 * u.MHz,
            "pol_type": "linear",
            "start_time": Time.now(),
        }

        x = p.exp(1j * p.random.uniform(-np.pi, +np.pi, shape))
        z = pb.DualPolarizationSignal(x, **kw)

        for n in [33, 32, shape[0]]:
            y = pb.contrib.istft(pb.contrib.stft(z, nperseg=n), nperseg=n)
            assert isinstance(y, type(z))
            assert isinstance(y.data, type(z.data))
            assert_equal_radiosignals(z[: len(y)], y)

    def test_single_tone(self):
        x = np.exp(2j * np.pi * np.arange(1024) * 0.25)[:, None]
        z = pb.BasebandSignal(x, sample_rate=1 * u.MHz, center_freq=1 * u.GHz)

        for n in [32, 64, 512, 1024]:
            y = pb.contrib.stft(z, nperseg=n)
            a = np.zeros_like(y)
            a[:, 3 * n // 4] = 1.0
            assert np.allclose(a, y.data)



"""Tests for `pulsarbat.utils`"""

import pytest
import types
import numpy as np
import dask.array as da
import scipy.fft
import pulsarbat as pb

FFT_FUNCS = [
    "fft",
    "fft2",
    "fftn",
    "ifft",
    "ifft2",
    "ifftn",
    "hfft",
    "ihfft",
    "rfft",
    "rfft2",
    "rfftn",
    "irfft",
    "irfft2",
    "irfftn",
]


class TestFFT:
    def test_dir_funcs(self):
        actual_dir = dir(pb.fft)
        expected_dir = sorted(FFT_FUNCS)

        for a, b in zip(actual_dir, expected_dir):
            assert a == b
            f = getattr(pb.fft, a)
            assert isinstance(f, types.FunctionType)

        with pytest.raises(AttributeError):
            _ = pb.fft.fish

    @pytest.mark.parametrize(
        "fft_func, N", [("fft", 8), ("ifft", 8), ("irfft", 9), ("hfft", 9)]
    )
    def test_complex_input_fft(self, fft_func, N):
        pb_fft_func = getattr(pb.fft, fft_func)
        sp_fft_func = getattr(scipy.fft, fft_func)

        for d in [np.float32, np.float64]:
            a = np.arange(N, dtype=d) + 1j * np.arange(N, dtype=d)
            x = sp_fft_func(a)

            y = pb_fft_func(a)
            assert type(a) == type(y)
            assert x.dtype == y.dtype
            assert np.allclose(x, y)

            b = da.arange(N, dtype=d) + 1j * da.arange(N, dtype=d)

            z = pb_fft_func(b)
            assert type(b) == type(z)
            assert x.dtype == z.dtype
            assert np.allclose(x, np.array(z))

    @pytest.mark.parametrize("fft_func, N", [("rfft", 16), ("ihfft", 16)])
    def test_real_input_fft(self, fft_func, N):
        pb_fft_func = getattr(pb.fft, fft_func)
        sp_fft_func = getattr(scipy.fft, fft_func)

        for d in [np.float32, np.float64]:
            a = np.arange(N, dtype=d)
            x = sp_fft_func(a)

            y = pb_fft_func(a)
            assert type(a) == type(y)
            assert x.dtype == y.dtype
            assert np.allclose(x, y)

            b = da.arange(N, dtype=d)

            z = pb_fft_func(b)
            assert type(b) == type(z)
            assert x.dtype == z.dtype
            assert np.allclose(x, np.array(z))



"""Provide a Phase class with integer and fractional part.

Written by Marten H. van Kerkwijk (@mhvk) for gh:mhvk/baseband_tasks.
Licensed under the GPLv3.
"""

import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, Longitude
from astropy.time.utils import two_sum, two_product


__all__ = ["Phase", "FractionalPhase"]


FRACTION_UFUNCS = {np.cos, np.sin, np.tan, np.spacing}

COMPARISON_UFUNCS = {
    np.equal,
    np.not_equal,
    np.less,
    np.less_equal,
    np.greater,
    np.greater_equal,
}


def day_frac(val1, val2, factor=None, divisor=None):
    """Return the sum of ``val1`` and ``val2`` as two float64s.
    The returned floats are an integer part and the fractional remainder,
    with the latter guaranteed to be within -0.5 and 0.5 (inclusive on
    either side, as the integer is rounded to even).
    The arithmetic is all done with exact floating point operations so no
    precision is lost to rounding error.  It is assumed the sum is less
    than about 1e16, otherwise the remainder will be greater than 1.0.
    Parameters
    ----------
    val1, val2 : array of float
        Values to be summed.
    factor : float, optional
        If given, multiply the sum by it.
    divisor : float, optional
        If given, divide the sum by it.
    Returns
    -------
    day, frac : float64
        Integer and fractional part of val1 + val2.
    """
    # Add val1 and val2 exactly, returning the result as two float64s.
    # The first is the approximate sum (with some floating point error)
    # and the second is the error of the float64 sum.
    sum12, err12 = two_sum(val1, val2)

    if factor is not None:
        sum12, carry = two_product(sum12, factor)
        carry += err12 * factor
        sum12, err12 = two_sum(sum12, carry)

    if divisor is not None:
        q1 = sum12 / divisor
        p1, p2 = two_product(q1, divisor)
        d1, d2 = two_sum(sum12, -p1)
        d2 += err12
        d2 -= p2
        q2 = (d1 + d2) / divisor  # 3-part float fine here; nothing can be lost
        sum12, err12 = two_sum(q1, q2)

    # get integer fraction
    day = np.floor(sum12 + 0.5)
    extra, frac = two_sum(sum12, -day)
    frac += extra + err12
    # Our fraction can now have gotten >0.5 or <-0.5, which means we would
    # loose one bit of precision. So, correct for that.
    excess = np.floor(frac + 0.5)
    day += excess
    extra, frac = two_sum(sum12, -day)
    frac += extra + err12
    return day, frac


def _parse_string(s):
    """Parse a string into two doubles.

    Generally, first part before the decimal dot, second after,
    but takes account of possible exponents.
    """
    s = s.strip().lower().translate(s.maketrans("d", "e"))
    if s[-1] == "j":
        s = s[:-1]
        factor = 1j
    else:
        factor = 1
    if s[0] == "+":
        s = s[1:]
    elif s[0] == "-":
        s = s[1:]
        factor *= -1

    # Just test string is basically OK (and reference value below).
    test = float(s) * factor

    s_float, exp, s_exp = s.partition("e")
    s_count, sep, s_frac = s_float.rpartition(".")
    if exp:
        exponent = int(s_exp)
        if exponent < 0:
            n = min(len(s_count), -exponent)
            s_frac = s_count[-n:] + s_frac
            s_count = s_count[:-n]
            exponent += n
        elif exponent > 0:
            n = min(len(s_frac), exponent)
            s_count = s_count + s_frac[:n]
            s_frac = s_frac[n:]
            exponent -= n
        factor *= 10 ** exponent

    frac = float("0." + s_frac) * factor
    count = float("0" + s_count) * factor

    assert count + frac == test
    return count, frac


_parse_strings = np.vectorize(_parse_string, otypes=[complex, complex])


class FractionalPhase(Longitude):
    """Phase without the cycle count, i.e., with a range of 1 cycle.

    This subclass of `~astropy.coordinates.Longitude` differs from it
    mostly in being able to take the fractional part of any
    `~baseband_tasks.phases.Phase` input.

    Parameters
    ----------
    angle : array, list, scalar, `~astropy.units.Quantity`,
        :class:`~astropy.coordinates.Angle` The angle value(s). If a tuple,
        will be interpreted as ``(h, m s)`` or ``(d, m, s)`` depending
        on ``unit``. If a string, it will be interpreted following the
        rules described for :class:`~astropy.coordinates.Angle`.
    unit : :class:`~astropy.units.UnitBase`, str, optional
        The unit of the value specified for the angle.  This may be any
        string that `~astropy.units.Unit` understands.  Must be an angular
        unit.  Default is 'cycle'.
    wrap_angle : :class:`~astropy.coordinates.Angle` or equivalent, optional
        Angle at which to wrap back to ``wrap_angle - 1 cycle``.
        If ``None`` (default), it will be taken to be 0.5 cycle unless
        ``angle`` has a ``wrap_angle`` attribute.

    Raises
    ------
    `~astropy.units.UnitsError`
        If a unit is not provided or it is not an angular unit.
    `TypeError`
        If the angle parameter is a :class:`~astropy.coordinates.Latitude`.
    """

    _default_wrap_angle = Angle(0.5, u.cycle)
    _equivalent_unit = _default_unit = u.cycle

    def __new__(cls, angle, unit=None, wrap_angle=None, **kwargs):
        # TODO: ideally, the Longitude/Angle/Quantity initializer by
        # default tries to convert to float also for structured arrays,
        # maybe via astype.
        if isinstance(angle, Phase):
            angle = angle["frac"]
        return super().__new__(cls, angle, unit=unit, wrap_angle=wrap_angle, **kwargs)


def check_imaginary(a):
    """Check whether a value is purely imaginary or purely real.

    Parameters
    ----------
    a : array_like or `~numpy.ndarray` subclass

    Returns
    -------
    real, imaginary : array of float, bool
        A float array, either just the input array if not complex
        or the real or imaginary part if complex, with the bool indicating
        whether it is the imaginary part.

    Raises
    ------
    ValueError
        If the input is complex and is not purely real or purely imaginary.
    """
    if np.iscomplexobj(a):
        if np.all(a.real == 0):
            return a.imag, True
        elif np.all(a.imag == 0):
            return a.real, False
        else:
            raise ValueError("cannot have mixed real/imaginary Phase")
    else:
        return a, False


class Phase(Angle):
    """Represent two-part phase.

    With one part the integer cycle count and the other the fractional phase.

    The class is a high-precision version of `~astropy.coordinates.Angle`
    that aims to behave just like it, but use its precision in operations
    where possible -- decaying to a `~astropy.units.Quantity` otherwise.

    For instance, addition and subtraction preserve precision, as do
    multiplication and division with dimensionless quantities.  Similarly,
    trigonometric functions use just the fractional phase.

    The phase can either be purely real or purely imaginary, not mixed.  If
    imaginary, using it in `~numpy.exp` will again preserve precision.

    Parameters
    ----------
    phase1, phase2 : array or `~astropy.units.Quantity`
        Two-part phase.  If not quantities, the assumed units are cycles.
    copy : bool, optional
        Ensure a copy is made.  Only relevant if ``phase1`` is a `Phase`
        and ``phase2`` is not given.
    subok : bool, optional
        If `False` (default), the returned array will be forced to be a
        `Phase`.  Otherwise, `Phase` subclasses will be passed through.
        Only relevant if ``phase1`` or ``phase2`` is a `Phase` subclass.

    Notes
    -----
    The machinery to keep precision is not complete; in particular, reductions
    such as summing along an axis will currently lose precision.

    Strings passed in to ``phase1`` or ``phase2`` are first converted to
    standard doubles, which may lead to loss of precision.  For long strings,
    use the `~baseband_tasks.phases.Phase.from_string` class method instead.
    """

    _equivalent_unit = _unit = _default_unit = u.cycle
    _phase_dtype = np.dtype({"names": ["int", "frac"], "formats": ["f8"] * 2})

    def __new__(cls, phase1, phase2=None, copy=True, subok=False):
        if isinstance(phase1, Phase):
            if phase2 is not None:
                phase1 = phase1 + phase2
                copy = False
            if not subok and type(phase1) is not cls:
                phase1 = phase1.view(cls)
            return phase1.copy() if copy else phase1

        phase1 = Angle(phase1, cls._unit, copy=False)

        if phase2 is not None:
            if isinstance(phase2, Phase):
                phase2 = phase2 + phase1
                if not subok and type(phase2) is not cls:
                    phase2 = phase2.view(cls)
                return phase2

            phase2 = Angle(phase2, cls._unit, copy=False)

        return cls.from_angles(phase1, phase2)

    @classmethod
    def from_angles(cls, phase1, phase2=None, factor=None, divisor=None, out=None):
        """Create a Phase instance from two angles.

        The two angles will be added, and possibly multiplied by a factor or
        divided by a divisor, preserving precision using two doubles, one for
        the integer part and one for the remainder.

        Note that this class method is mostly meant for internal use.

        Parameters
        ----------
        phase1 : `~astropy.units.Quantity`
             With angular units.
        phase2 : `~astropy.units.Quantity` or `None`
             With angular units.
        factor : float or complex
             Posisble factor to multiply the angles with.
        divisor : float or complex
             Posisble divisor to divide the angles by.

        Raises
        ------
        ValueError
            If the result is not purely real or purely imaginary
        """
        phase1, imaginary = check_imaginary(phase1)
        if phase2 is not None:
            phase2, im2 = check_imaginary(phase2)
            if im2 is not imaginary:
                raise ValueError(
                    "phase1 and phase2 must either be both " "real or both imaginary."
                )
        if factor is not None:
            factor, imf = check_imaginary(factor)
            imaginary ^= imf
        if divisor is not None:
            divisor, imd = check_imaginary(divisor)
            if imd and not imaginary:
                divisor = -divisor
            imaginary ^= imd

        # TODO: would be nice if day_frac had an out parameter.
        phase1_value = phase1.to_value(cls._unit)
        if phase2 is None:
            phase2_value = 0.0
        else:
            phase2_value = phase2.to_value(cls._unit)
        count, fraction = day_frac(
            phase1_value, phase2_value, factor=factor, divisor=divisor
        )
        if out is None:
            value = np.empty(count.shape, cls._phase_dtype)
            out = value.view(cls)
        else:
            value = out.view(np.ndarray)
        value["int"] = count
        value["frac"] = fraction
        out.imaginary = imaginary
        return out

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        self.imaginary = getattr(obj, "imaginary", False)

    def __getitem__(self, item):
        result = super().__getitem__(item)
        if result.dtype is self.dtype:
            return result

        if item == "frac":
            result = result.view(Angle if self.imaginary else FractionalPhase)
        else:
            assert item == "int"
            result = result.view(Angle)

        if self.imaginary:
            result = result * 1j

        return result

    def __iter__(self):
        if self.isscalar:
            raise TypeError(
                "'{cls}' object with a scalar value is not iterable".format(
                    cls=self.__class__.__name__
                )
            )

        # Override Quantity.__iter__ since that iterates over self.value.
        for i in range(len(self)):
            yield self[i]

    def _set_unit(self, unit):
        if unit is None or unit != self._unit:
            raise u.UnitTypeError(
                "{0} instances require units of '{1}'".format(
                    type(self).__name__, self._unit
                )
                + (
                    ", but no unit was given."
                    if unit is None
                    else ", so cannot set it to '{0}'.".format(unit)
                )
            )

        super()._set_unit(unit)

    def __repr__(self):
        v = self.view(np.ndarray)
        return "{0}({1}{3} cycle, {2}{3} cycle)".format(
            self.__class__.__name__,
            v["int"],
            v["frac"],
            " * 1j" if self.imaginary else "",
        )

    def __str__(self):
        # Override Angle, since one cannot override the formatter for
        # structured dtype in array2string.
        def formatter(x):
            # [...] and view as self.__class__ are needed for scalar case.
            return x[...].view(self.dtype, self.__class__).to_string()

        return (
            np.array2string(
                self.view(f"V{self.dtype.itemsize}"), formatter={"all": formatter}
            )
            + self._unitstr
        )

    def __format__(self, format_spec):
        """Format a phase, special-casing the float format.

        For the 'f' format, precision is kept, and the unit is suppressed.
        For everything else, the Quantity formatter is used.
        """
        if format_spec.endswith("f"):
            # Check that formatting works at all...
            test = format(self.value, format_spec)
            pre, dot, post = test.partition(".")
            if post:
                precise = self.to_string(precision=len(post))
                pre, _, post = precise.partition(".")
                # Just to ensure no bad rounding happened
                pre = format(float(pre), format_spec).partition(".")[0]
                return pre + dot + post

        return self.cycle.__format__(format_spec)

    def to_string(
        self,
        unit=None,
        decimal=True,
        sep="fromunit",
        precision=None,
        alwayssign=False,
        pad=False,
        fields=3,
        format=None,
    ):
        """ A string representation of the Phase.

        By default, uses a decimal representation that is guaranteed to
        preserve precision to within 1e-16 cycles.  Otherwise, uses
        `astropy.coordinates.Angle.to_string`.
        """
        if not decimal or (unit is not None and unit != u.cycle):
            return self.cycle.to_string(
                unit=unit,
                decimal=decimal,
                sep=sep,
                precision=precision,
                alwayssign=alwayssign,
                pad=pad,
                fields=fields,
                format=format,
            )

        if precision is None:
            func = str
        else:
            func = ("{0:1." + str(precision) + "f}").format

        def do_format(count, frac):
            neg = (count + frac) < 0
            if neg:
                count = -count
                frac = -frac
                sign = "-"
            elif alwayssign:
                sign = "+"
            else:
                sign = ""

            if frac < 0:
                frac += 1
                count -= 1

            if frac < 0.25:
                # Ensure that we do not get 1e-16, etc., yet can use numpy's
                # guarantee that the right number of digits is shown.
                frac_str = func(frac + 0.25)
                f24 = int(frac_str[2:4])
                if func is str and (
                    len(frac_str) == 3 or (len(frac_str) == 4 and frac_str[3] == "5")
                ):
                    if len(frac_str) == 3:
                        f24 = "{:02d}".format(f24 * 10 - 25)
                    else:
                        f24 = "{:1d}".format((f24 - 5) // 10 - 2)
                else:
                    f24 = "{:02d}".format(f24 - 25)
                frac_str = frac_str[:2] + f24 + frac_str[4:]
            else:
                frac_str = func(frac)
                if frac_str[0] == "1":
                    count += 1
            s = sign + str(int(count)) + frac_str[1:]
            if self.imaginary:
                s += "j"
            if format == "latex":
                s = "$" + s + "$"
            return s

        format_ufunc = np.vectorize(do_format, otypes=["U"])
        if self.imaginary:
            count, frac = self["int"].value.imag, self["frac"].value.imag
        else:
            count, frac = self["int"].value, self["frac"].value

        result = format_ufunc(count, frac)

        if result.ndim == 0:
            result = result[()]
        return result

    @classmethod
    def from_string(cls, string):
        """Create Phase instance from a long string.

        The string has to be a standard decimal string, i.e., no attempt is
        made to parse an angle.
        """
        string = np.asanyarray(string)
        if string.dtype.kind not in "SU":
            raise ValueError("require string input.")
        count, frac = _parse_strings(string)
        return cls(count, frac)

    @property
    def int(self):
        """Rounded cycle count."""
        return self["int"]

    @property
    def frac(self):
        """Fractional phase, between -0.5 and 0.5 cycles."""
        return self["frac"]

    @property
    def cycle(self):
        """Full cycle, including phase, as a regular Angle.

        The result will use a standard double, and thus likely loose precision.
        """
        return self["int"] + self["frac"]

    def to_value(self, unit=None, equivalencies=[]):
        """The numerical value, possibly in a different unit.

        The result will use a standard double, and thus likely loose precision.
        """
        return self.cycle.to_value(unit, equivalencies)

    value = property(
        to_value,
        doc="""The numerical value.

    The result will use a standard double, and thus likely loose precision.

    See also
    --------
    to_value : Get the numerical value in a given unit.
    """,
    )

    def to(self, unit, equivalencies=[]):
        """The phase in a different unit.

        For any unit except "cycle", this will likely loose precision as
        an `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        is returned.
        """
        if unit == u.cycle:
            return self.copy()
        else:
            return self.cycle.to(unit, equivalencies=equivalencies)

    def _take_along_axis(self, indices, axis=None, keepdims=False):
        # More or less a straight copy from Time.
        if axis is None:
            return self[np.unravel_index(indices, self.shape)]

        if indices.ndim == self.ndim - 1:
            indices = np.expand_dims(indices, axis)

        result = np.take_along_axis(self, indices, axis)

        return result if keepdims else result.squeeze(axis)

    def argmin(self, axis=None, out=None):
        """Return indices of the minimum values along the given axis."""
        approx = np.min(self.cycle, axis, keepdims=True)
        dt = (self["int"] - approx) + self["frac"]
        return dt.argmin(axis, out)

    def argmax(self, axis=None, out=None):
        """Return indices of the maximum values along the given axis."""
        approx = np.max(self.cycle, axis, keepdims=True)
        dt = (self["int"] - approx) + self["frac"]
        return dt.argmax(axis, out)

    def argsort(self, axis=-1):
        """Returns the indices that would sort the phase array."""
        phase_approx = self.cycle
        phase_remainder = (self - phase_approx).cycle
        if axis is None:
            return np.lexsort((phase_remainder.ravel(), phase_approx.ravel()))
        else:
            return np.lexsort(keys=(phase_remainder, phase_approx), axis=axis)

    # Below are basically straight copies from Time
    def min(self, axis=None, out=None, keepdims=False):
        """Minimum along a given axis.

        This is similar to :meth:`~numpy.ndarray.min`, but adapted to ensure
        that the full precision is used.
        """
        if out is not None:
            raise ValueError("An `out` argument is not yet supported.")
        return self._take_along_axis(self.argmin(axis), axis, keepdims)

    def max(self, axis=None, out=None, keepdims=False):
        """Maximum along a given axis.

        This is similar to :meth:`~numpy.ndarray.max`, but adapted to ensure
        that the full precision is used.
        """
        if out is not None:
            raise ValueError("An `out` argument is not yet supported.")
        return self._take_along_axis(self.argmax(axis), axis, keepdims)

    def ptp(self, axis=None, out=None, keepdims=False):
        """Peak to peak (maximum - minimum) along a given axis.

        This is similar to :meth:`~numpy.ndarray.ptp`, but adapted to ensure
        that the full precision is used.
        """
        if out is not None:
            raise ValueError("An `out` argument is not yet supported.")
        return self.max(axis, keepdims=keepdims) - self.min(axis, keepdims=keepdims)

    def sort(self, axis=-1):
        """Return a copy sorted along the specified axis.

        This is similar to :meth:`~numpy.ndarray.sort`, but internally uses
        indexing with :func:`~numpy.lexsort` to ensure that the full precision
        given by the two doubles is kept.

        Parameters
        ----------
        axis : int or None
            Axis to be sorted.  If ``None``, the flattened array is sorted.
            By default, sort over the last axis.
        """
        return self._take_along_axis(self.argsort(axis), axis, keepdims=True)

    # Quantity lets ndarray.__eq__, __ne__ deal with structured arrays like us.
    # Override this so we can deal with it in __array_ufunc__.
    def __eq__(self, other):
        try:
            return np.equal(self, other)
        except u.UnitsError:
            return False
        except Exception:
            return NotImplemented

    def __ne__(self, other):
        try:
            return np.not_equal(self, other)
        except u.UnitsError:
            return True
        except Exception:
            return NotImplemented

    def __array_ufunc__(self, function, method, *inputs, **kwargs):
        """Wrap numpy ufuncs, taking care of units.

        Parameters
        ----------
        function : callable
            ufunc to wrap.
        method : str
            Ufunc method: ``__call__``, ``at``, ``reduce``, etc.
        inputs : tuple
            Input arrays.
        kwargs : keyword arguments
            As passed on, with ``out`` containing possible quantity output.

        Returns
        -------
        result : array_like
            Results of the ufunc, with the unit set properly,
            `~baseband_tasks.phases.Phase` if possible (i.e., with units of
            cycles), otherwise `~astropyy.units.Quantity` or `~numpy.ndarray`
            as appropriate.
        """
        # Do *not* use inputs.index(self) since that will use __eq__
        for i_self, input_ in enumerate(inputs):
            if input_ is self:
                break
        else:
            i_self += 1

        # Some bools for use in the if-statements below.
        # TODO: support reductions on add, minimum, maximum; others?
        basic = method == "__call__" and i_self < function.nin
        basic_real = basic and not self.imaginary
        basic_phase_out = basic

        out = kwargs.get("out", None)
        if out is not None:
            if len(out) == 1 and isinstance(out[0], Phase):
                phase_out = out[0]
                out = None
            else:
                phase_out = None
                basic_phase_out = False
        else:
            phase_out = None

        if function in {np.add, np.subtract} and basic and out is None:
            try:
                phases = [Phase(input_, copy=False, subok=True) for input_ in inputs]
            except Exception:
                return NotImplemented

            if phases[0].imaginary == phases[1].imaginary:
                return self.from_angles(
                    function(phases[0]["int"], phases[1]["int"]),
                    function(phases[0]["frac"], phases[1]["frac"]),
                    out=phase_out,
                )

        elif function in COMPARISON_UFUNCS and basic:
            phases = list(inputs)
            try:
                phases[1 - i_self] = Phase(inputs[1 - i_self], copy=False, subok=True)
            except Exception:
                return NotImplemented

            if phases[0].imaginary == phases[1].imaginary:
                diff = (phases[0]["int"] - phases[1]["int"]) + (
                    phases[0]["frac"] - phases[1]["frac"]
                )
                return getattr(function, method)(diff, 0, **kwargs)

        elif (
            function is np.multiply or function is np.divide and i_self == 0
        ) and basic_phase_out:
            try:
                other = u.Quantity(
                    inputs[1 - i_self], u.dimensionless_unscaled, copy=False
                ).value
                if function is np.multiply:
                    return self.from_angles(
                        self["int"], self["frac"], factor=other, out=phase_out
                    )
                else:
                    return self.from_angles(
                        self["int"], self["frac"], divisor=other, out=phase_out
                    )
            except Exception:
                # If not consistent with a dimensionless quantity, or mixed
                # real and complex, we follow the standard route of
                # downgrading ourself to a quantity and see if things work.
                pass

        elif function in {np.floor_divide, np.remainder, np.divmod} and basic_real:
            fd_out = None
            if out is not None:
                if function is np.divmod:
                    fd_out = out[0]
                    phase_out = out[1]
                elif function is np.floor_divide:
                    fd_out = out[0]
            elif phase_out is not None and function is np.floor_divide:
                return NotImplemented

            fd = np.floor_divide(self.cycle, inputs[1], out=fd_out)
            corr = Phase.from_angles(inputs[1], factor=fd, out=phase_out)
            remainder = np.subtract(self, corr, out=corr)
            fdx = np.floor_divide(remainder.cycle, inputs[1])
            # This can likely be optimized...
            # Note: one cannot just loop, because rounding of exact 0.5.
            # TODO: check this method is really correct.
            if np.count_nonzero(fdx):
                fd += fdx
                corr = Phase.from_angles(inputs[1], factor=fd, out=corr)
                remainder = np.subtract(self, corr, out=corr)

            if function is np.floor_divide:
                return fd
            elif function is np.remainder:
                return remainder
            else:
                return fd, remainder

        elif function is np.positive and basic_phase_out:
            return self.from_angles(self["int"], self["frac"], out=phase_out)

        elif function is np.negative and basic_phase_out:
            return self.from_angles(-self["int"], -self["frac"], out=phase_out)

        elif function in {np.absolute, np.fabs} and basic_phase_out:
            # Go via view to avoid having to deal with imaginary.
            v = self.view(np.ndarray)
            return self.from_angles(
                u.Quantity(v["int"], u.cycle, copy=False),
                u.Quantity(v["frac"], u.cycle, copy=False),
                factor=np.sign(v["int"] + v["frac"]),
                out=phase_out,
            )

        elif function is np.rint and basic:
            return np.positive(self["int"], **kwargs)

        elif function in FRACTION_UFUNCS and basic_real:
            frac = self.frac.view(Angle)
            return function(frac, **kwargs)

        elif function is np.exp and basic and self.imaginary:
            # Avoid dimensionless_angles, but still get Quantity out.
            exponent = u.Quantity(self.frac.to_value(u.radian), copy=False)
            return function(exponent, **kwargs)

        # Fall-back: treat Phase as a simple Quantity.
        if basic:
            inputs = tuple(
                (input_.cycle if isinstance(input_, Phase) else input_)
                for input_ in inputs
            )
            quantity = inputs[i_self]
        else:
            quantity = self.cycle

        if phase_out is None:
            return quantity.__array_ufunc__(function, method, *inputs, **kwargs)
        else:
            # We won't be able to store in a phase directly, but might
            # as well use one of its elements to store the angle.
            kwargs["out"] = (phase_out["int"],)
            result = quantity.__array_ufunc__(function, method, *inputs, **kwargs)
            return phase_out.from_angles(result, out=phase_out)

    def _new_view(self, obj=None, unit=None):
        # If the unit is not right, we should ensure we change our two-float
        # dtype to a single float.
        if unit is not None and unit != self._unit:
            if obj is None:
                obj = self.cycle
            elif isinstance(obj, Phase):
                obj = obj.cycle

        return super()._new_view(obj, unit)

    def astype(self, dtype, order="K", casting="unsafe", subok=True, copy=True):
        """Copy of the array, cast to a specified type.

        As `numpy.ndarray.astype`, but using knowledge of format to cast to
        floats.
        """
        dtype = np.dtype(dtype)
        if not dtype.fields and casting in {"same_kind", "unsafe"}:
            parts = [
                self[part].astype(
                    dtype, order=order, casting=casting, subok=subok, copy=copy
                )
                for part, copy in (("int", True), ("frac", False))
            ]
            parts[0] += parts[1]
            return parts[0]

        else:
            return super().astype(
                dtype, order=order, casting=casting, subok=subok, copy=copy
            )



"""Collection of handy utilities."""

import numpy as np
from functools import lru_cache
import scipy.fft


__all__ = [
    "real_to_complex",
    "next_fast_len",
    "prev_fast_len",
]


def real_to_complex(z, axis=0):
    """Convert a real baseband signal to a complex baseband signal.

    This function computes the analytic representation of the input
    signal via a Hilbert transform, throwing away negative frequency
    components. Then, the signal is shifted in frequency domain by -B/2
    where B is the bandwidth of the signal. Finally, the signal is
    decimated by a factor of 2, which results in the complex baseband
    representation of the input signal.

    Parameters
    ----------
    z : `~numpy.ndarray`
        Input signal. Must be real.
    axis : int, optional
        Axis over which to convert the signal. This will be the axis
        that represents time. Default is 0.

    Returns
    -------
    out : np.ndarray
        The complex baseband representation of the input signal.
    """
    z = np.asarray(z)
    if np.iscomplexobj(z):
        raise ValueError("Input must be real-valued.")

    out_dtype = np.complex64 if z.dtype == np.float32 else np.complex128
    N = z.shape[axis]

    if N == 0:
        return z.astype(out_dtype)

    # Pick the correct axis to work on
    ind = [np.newaxis] * z.ndim
    ind[axis] = slice(None)

    # Get analytic signal via Hilbert transform and shift by -B/2
    h = np.zeros(N, dtype=out_dtype)
    h[0] = 1
    h[1 : N // 2] = 2
    if N > 1:
        h[N // 2] = 2 if N % 2 else 1

    z = scipy.fft.ifft(scipy.fft.fft(z, axis=axis) * h[tuple(ind)], axis=axis)
    z *= np.exp(-1j * np.pi / 2 * np.arange(N))[tuple(ind)]

    # Decimate signal by factor of 2 (along axis)
    dec = [slice(None)] * z.ndim
    dec[axis] = slice(None, None, 2)
    return z[tuple(dec)].astype(out_dtype)


@lru_cache(maxsize=1024)
def next_fast_len(N):
    """Returns smallest 7-smooth number >= N."""
    if N <= 10:
        return N

    f7, guess = 1, 2 * N
    while f7 < guess:
        f75 = f7
        while f75 < guess:
            x = f75

            while x < N:
                x *= 2

            while 1:
                if x < N:
                    x *= 3
                elif x > N:
                    if x < guess:
                        guess = x
                    if x & 1:
                        break
                    x >>= 1
                else:
                    return N

            f75 *= 5
        f7 *= 7
    return guess


@lru_cache(maxsize=1024)
def prev_fast_len(N):
    """Returns largest 7-smooth number <= N."""
    if N <= 10:
        return N

    f7, guess = 1, 1
    while f7 <= N:
        f75 = f7
        while f75 <= N:
            x = f75

            while x <= N:
                x *= 2
            x >>= 1

            while 1:
                if x < N:
                    if x > guess:
                        guess = x
                    x *= 3
                elif x > N:
                    if x & 1:
                        break
                    x >>= 1
                else:
                    return N

            f75 *= 5
        f7 *= 7
    return guess



"""Tests for readers in `pulsarbat.reader`."""

import pytest
from pathlib import Path
import numpy as np
import astropy.units as u
from astropy.time import Time
import cloudpickle
import pulsarbat as pb
import pulsarbat.readers as pbr

DATA_DIR = Path(__file__).parent.absolute() / "data"


class TestBasebandReader:
    def test_vdif(self):
        r = pbr.BasebandReader(DATA_DIR / "sample.vdif")
        assert r.shape == (20000, 8)
        t0 = Time("2014-06-16T05:56:07.000", format="isot", precision=9)
        assert Time.isclose(r.start_time, t0)

        z = r.read(0, 16)
        assert u.isclose(z.sample_rate, 16 * u.MHz)

        r = pbr.BasebandReader(DATA_DIR / "sample.vdif", squeeze=False)
        assert r.shape == (20000, 8, 1)

        # Test picklability
        _ = cloudpickle.loads(cloudpickle.dumps(r))

    def test_dada(self):
        r = pbr.BasebandReader(DATA_DIR / "sample.dada")
        assert r.shape == (16000, 2)
        t0 = Time("2013-07-02T01:39:20.000", format="isot", precision=9)
        assert Time.isclose(r.start_time, t0)

        z = r.read(0, 16)
        assert u.isclose(z.sample_rate, 16 * u.MHz)

        r = pbr.BasebandReader(DATA_DIR / "sample.dada", squeeze=False)
        assert r.shape == (16000, 2, 1)

        # Test picklability
        _ = cloudpickle.loads(cloudpickle.dumps(r))

    def test_sideband(self):
        f = DATA_DIR / "sample.vdif"
        r1 = pbr.BasebandReader(f, lower_sideband=False)
        r2 = pbr.BasebandReader(f, lower_sideband=True)

        z1, z2 = r1.read(0, 16), r2.read(0, 16)
        assert np.allclose(np.array(z1), np.array(z2).conj())

        lsb = (np.arange(8) % 3).astype(bool)
        r3 = pbr.BasebandReader(f, lower_sideband=lsb)
        z3 = r3.read(0, 16)
        assert np.allclose(np.array(z1)[:, ~lsb], np.array(z3)[:, ~lsb])
        assert np.allclose(np.array(z1)[:, lsb], np.array(z3)[:, lsb].conj())

        with pytest.raises(ValueError):
            _ = pbr.BasebandReader(f, lower_sideband=[True, False])

    def test_intensity_contraints(self):
        with pytest.raises(ValueError):
            _ = pbr.BasebandReader(DATA_DIR / "sample.dada", intensity=True)

        sig_kw = dict(center_freq=1.4*u.GHz, chan_bw=16*u.MHz)

        r = pbr.BasebandReader(DATA_DIR / "sample.vdif", signal_kwargs=sig_kw,
                               signal_type=pb.IntensitySignal)
        assert r.intensity

        with pytest.raises(ValueError):
            _ = pbr.BasebandReader(DATA_DIR / "sample.vdif", signal_kwargs=sig_kw,
                                   signal_type=pb.IntensitySignal, intensity=False)

        sig_kw = dict(center_freq=1.4*u.GHz)

        r = pbr.BasebandReader(DATA_DIR / "sample.vdif", signal_kwargs=sig_kw,
                               signal_type=pb.BasebandSignal)
        assert not r.intensity

        with pytest.raises(ValueError):
            _ = r = pbr.BasebandReader(DATA_DIR / "sample.vdif", signal_kwargs=sig_kw,
                                       signal_type=pb.BasebandSignal, intensity=True)


class TestGUPPIRawReader:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_basic(self, use_dask):
        fs = sorted(DATA_DIR.glob("fake.*.raw"))
        r = pbr.GUPPIRawReader(fs)
        assert len(r) == 8192 * len(fs)

        # Test picklability
        _ = cloudpickle.loads(cloudpickle.dumps(r))

        assert u.isclose(r.center_freq, 344.1875 * u.MHz)
        assert r.pol_type == "linear"

        z = r.read(0, 8, use_dask=use_dask)
        assert len(z) == 8
        assert u.isclose(z.sample_rate, 3.125 * u.MHz)
        assert u.isclose(z.center_freq, 344.1875 * u.MHz)
        assert u.isclose(z.bandwidth, 12.5 * u.MHz)

        t0 = Time("1997-07-11T12:34:56.000", format="isot", precision=9)
        assert Time.isclose(z.start_time, t0)
        assert z.pol_type == "linear"

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_sequential(self, use_dask):
        fs = sorted(DATA_DIR.glob("fake.*.raw"))

        r1 = pbr.GUPPIRawReader(fs)
        z1 = r1.read(16384 + 16, 32, use_dask=use_dask)

        r2 = pbr.GUPPIRawReader(fs[2])
        z2 = r2.read(16, 32, use_dask=use_dask)

        assert np.allclose(np.array(z1), np.array(z2))
        assert Time.isclose(z1.start_time, z2.start_time)


class TestDADAStokesReader:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_basic(self, use_dask):
        r = pbr.DADAStokesReader(DATA_DIR / "stokes_ef.dada")
        assert r.shape == (16, 2048, 4)

        z = r.read(0, 4, use_dask=use_dask)
        assert type(z) == pb.FullStokesSignal
        assert u.isclose(z.center_freq, 7 * u.GHz)
        assert u.isclose(z.bandwidth, 2 * u.GHz)
        assert u.isclose(z.dt, 131072 * u.ns)
        assert z.freq_align == "top"
        assert z.nchan == 2048

    def test_not_stokes(self):
        with pytest.raises(ValueError):
            _ = pbr.DADAStokesReader(DATA_DIR / "sample.dada")



"""Pulsar processing module."""

# flake8: noqa

from .phase import FractionalPhase, Phase
from .predictor import PolycoEntry, PhasePredictor

__all__ = ["PolycoEntry", "PhasePredictor", "Phase", "FractionalPhase"]



"""FFT Functions with proper dispatching for Dask Arrays."""

import scipy.fft
from functools import singledispatch
import dask.array as da


_FFT_FUNCS = [
    "fft",
    "fft2",
    "fftn",
    "ifft",
    "ifft2",
    "ifftn",
    "rfft",
    "rfft2",
    "rfftn",
    "irfft",
    "irfft2",
    "irfftn",
    "hfft",
    "ihfft",
]


def __dir__():
    return sorted(_FFT_FUNCS)


def __getattr__(name):
    if name not in _FFT_FUNCS:
        raise AttributeError(f"module {__name__} has no attribute {name}")

    _fft_func = getattr(scipy.fft, name)

    @singledispatch
    def func(*args, **kwargs):
        return _fft_func(*args, **kwargs)

    @func.register(da.Array)
    def _(*args, **kwargs):
        wrapped_func = da.fft.fft_wrap(_fft_func)
        return wrapped_func(*args, **kwargs)

    func.__qualname__ = _fft_func.__qualname__
    func.__name__ = _fft_func.__name__
    func.__doc__ = _fft_func.__doc__
    return func



"""Tests for `pulsarbat.Signal` class."""

import pytest
import numpy as np
import dask.array as da
import astropy.units as u
from astropy.time import Time
import pulsarbat as pb
import cloudpickle

RAND = np.random.default_rng(seed=42)


@pytest.mark.parametrize("fn", [np.random.standard_normal, da.random.standard_normal])
@pytest.mark.parametrize("shape", [(16,), (8, 4)])
def test_signal_basic(fn, shape):
    for x in [fn(shape), fn(shape) + 1j * fn(shape)]:
        z = pb.Signal(x, sample_rate=1 * u.Hz)

        # Test picklability
        _ = cloudpickle.loads(cloudpickle.dumps(z))

        assert isinstance(z, pb.Signal)
        assert z.shape == shape
        assert z.ndim == len(shape)
        assert len(z) == shape[0]
        assert z.sample_shape == shape[1:]

        assert z.dtype == x.dtype
        assert type(z.data) is type(x)
        assert np.all(np.array(z) == np.array(x))


@pytest.mark.parametrize("ts", [None, Time.now()])
@pytest.mark.parametrize("sample_rate", [1 * u.Hz, 33 * u.kHz])
@pytest.mark.parametrize("meta", [None, {"name": "Hello!"}])
def test_signal_attrs(ts, sample_rate, meta):
    shape = (8, 2)
    z = pb.Signal(np.ones(shape), sample_rate=sample_rate, start_time=ts, meta=meta)

    print(z)
    repr(z)

    assert z.sample_rate == sample_rate
    assert u.isclose(z.dt, 1 / sample_rate)
    assert u.isclose(z.time_length, shape[0] / sample_rate)

    if ts is None:
        assert z.start_time is None
        assert z.stop_time is None
        assert Time.now() not in z
    else:
        assert z.start_time == ts
        assert z.stop_time == ts + (shape[0] / sample_rate)
        assert z.start_time in z
        assert z.stop_time not in z
        assert ts + 0.1 * u.us in z


def test_time_contains():
    def check(n, a, b):
        x = np.ones(n, dtype=bool)
        x[:a] = False
        x[-b:] = False
        return x

    t0 = Time("2018-08-01T13:56:23", format="isot", precision=9)
    z = pb.Signal(np.zeros(10), sample_rate=1 * u.Hz, start_time=t0)

    ts = t0 + np.arange(-2, 12) * z.dt
    k = check(len(ts), 2, 2)
    assert np.allclose(k, z.contains(ts))
    assert np.allclose(k, [t in z for t in ts])

    ts -= 1 * u.ns
    k = check(len(ts), 3, 1)
    assert np.allclose(k, z.contains(ts))
    assert np.allclose(k, [t in z for t in ts])


def test_meta_errors():
    for meta in ["Hello", (0,), {0, 4}]:
        with pytest.raises(ValueError):
            _ = pb.Signal(np.empty((16, 4)), sample_rate=1 * u.Hz, meta=meta)


def test_shape_errors():
    for shape in [(), (15, 0, 4), (4, 0)]:
        with pytest.raises(ValueError):
            _ = pb.Signal(np.empty(shape), sample_rate=1 * u.Hz)


def test_sample_rate_errors():
    for x in [99.0, 4 * u.m, [5, 6] * u.Hz, -1 * u.Hz]:
        with pytest.raises(ValueError):
            _ = pb.Signal(np.empty((8, 4)), sample_rate=x)


@pytest.mark.parametrize(
    "ts", [Time([59843, 45678, 65678], format="mjd"), 59867.2442234]
)
def test_start_time_errors(ts):
    with pytest.raises(ValueError):
        _ = pb.Signal(np.empty((8, 4)), sample_rate=1 * u.Hz, start_time=ts)


class ArbitrarySignal(pb.Signal):
    _req_dtype = (np.int16, np.int32)
    _req_shape = (None, 7, 4, None)
    _axes_labels = {"time": 0, "fish": 1, "foo": 2, "bar": 3}

    def __init__(self, z, /, *, sample_rate, foo):
        self._foo = foo
        super().__init__(z, sample_rate=sample_rate)

    @property
    def foo(self):
        return self._foo


def test_dtype_constraints():
    shape = (10, 7, 4, 1)

    for dtype in [np.int8, np.int16]:
        x = np.zeros(shape, dtype=dtype)
        z = ArbitrarySignal(x, sample_rate=1 * u.Hz, foo="bar")

        # Test picklability
        _ = cloudpickle.loads(cloudpickle.dumps(z))

        assert z.dtype == np.int16

    for dtype in [np.float32, np.int64]:
        x = np.empty(shape, dtype=dtype)
        with pytest.raises(ValueError):
            _ = ArbitrarySignal(x, sample_rate=1 * u.Hz, foo="bar")


def test_shape_constraints():
    for shape in [(), (0, 7, 4), (11, 7, 4, 0), (3, 6, 5, 1)]:
        x = np.empty(shape).astype(np.int8)
        with pytest.raises(ValueError):
            _ = ArbitrarySignal(x, sample_rate=1 * u.Hz, foo="bar")


def test_signal_like():
    x1 = RAND.integers(-50, 50, (10, 7, 4, 1)).astype(np.int8)

    z1 = ArbitrarySignal(x1, sample_rate=1 * u.Hz, foo="bar")
    z2 = pb.Signal.like(z1)

    assert np.array_equal(np.array(z1), np.array(z2))
    assert z1.sample_rate == z2.sample_rate
    assert z1.start_time == z2.start_time
    assert type(z2) is pb.Signal

    x2 = RAND.integers(-50, 50, (10, 7, 4, 1)).astype(np.int8)
    z2 = pb.Signal.like(z1, x2)
    assert not np.array_equal(np.array(z1), np.array(z2))

    z2 = pb.Signal.like(z1, sample_rate=2 * z1.sample_rate)
    assert u.isclose(z2.sample_rate, 2 * z1.sample_rate)

    with pytest.raises(ValueError):
        _ = ArbitrarySignal.like(z2)

    z3 = ArbitrarySignal.like(z2, foo="bar")
    assert np.array_equal(np.array(z1), np.array(z3))
    assert z1.foo == z3.foo == "bar"
    assert u.isclose(z3.sample_rate, 2 * z1.sample_rate)


def test_signal_slice():
    x = np.ones((32, 4, 2), dtype=np.float32)
    z = pb.Signal(x, sample_rate=1 * u.Hz, start_time=Time.now())

    y = z[2:8]
    assert len(y) == 6
    Time.isclose(y.start_time - 2 * u.s, z.start_time)
    assert u.isclose(z.sample_rate, y.sample_rate)
    assert u.isclose(y.time_length, 6 * u.s)

    y = z[::4]
    assert len(y) == 8
    Time.isclose(y.start_time, z.start_time)
    assert u.isclose(y.sample_rate, 0.25 * u.Hz)
    assert u.isclose(z.time_length, y.time_length)

    y = z[:, [0, 2]]
    assert len(y) == len(z)
    Time.isclose(y.start_time, z.start_time)
    assert u.isclose(z.sample_rate, y.sample_rate)
    assert u.isclose(z.time_length, y.time_length)

    b = RAND.random((4, 2)) > 0.5
    y = z[:, b]
    assert len(y) == len(z)
    Time.isclose(y.start_time, z.start_time)
    assert u.isclose(z.sample_rate, y.sample_rate)
    assert u.isclose(z.time_length, y.time_length)

    y = z[5:3]
    assert len(y) == 0
    assert Time.isclose(y.start_time - 5 * u.s, z.start_time)
    assert Time.isclose(y.stop_time - 5 * u.s, z.start_time)
    assert u.isclose(y.time_length, 0 * u.s)

    z = pb.Signal(x, sample_rate=1 * u.Hz)
    assert z[2:].start_time is None
    assert z[:6].start_time is None
    assert z[::4].start_time is None


def test_signal_slice_errors():
    x = np.ones((32, 4, 2), dtype=np.float32)
    st = Time.now()
    z = pb.Signal(x, sample_rate=1 * u.Hz, start_time=st)

    with pytest.raises(IndexError):
        _ = z["key"]

    with pytest.raises(IndexError):
        _ = z[4]

    with pytest.raises(IndexError):
        b = np.random.random(len(z)) > 0.5
        _ = z[b]

    with pytest.raises(AssertionError):
        _ = z[::-2]


def test_get_axis():
    x = np.zeros((5, 7, 4, 3), dtype=np.int32)
    z = ArbitrarySignal(x, sample_rate=1 * u.Hz, foo="bar")

    for a in [-5, 4]:
        with pytest.raises(ValueError):
            _ = z.get_axis(a)

    for a in range(-4, 4):
        assert a == z.get_axis(a)

    for i, a in enumerate(["time", "fish", "foo", "bar"]):
        assert i == z.get_axis(a)


class TestSignalUfuncs:
    @pytest.mark.parametrize("use_dask", [True, False])
    def test_basic_ufunc(self, use_dask):
        arange = da.arange if use_dask else np.arange
        x = arange(8, dtype=np.int64) + 11
        y = arange(8, dtype=np.int64) + 2

        a = pb.Signal(x, sample_rate=2 * u.Hz)
        b = pb.Signal(y, sample_rate=5 * u.Hz)

        for s in [a + 2, a + y, 2 + a, y + a, a + b]:
            assert type(s) == type(a)
            assert s.sample_rate == a.sample_rate

        assert np.allclose(np.array(a + 2), np.array(x + 2))
        assert np.allclose(np.array(a + y), np.array(x + y))
        assert np.allclose(np.array(a + b), np.array(x + y))

        z = x + y
        s = a
        a += b
        assert np.allclose(np.array(a), np.array(z))
        assert np.allclose(np.array(s), np.array(z))

        empty = da.empty if use_dask else np.empty
        c = pb.Signal(empty(a.shape), sample_rate=10 * u.Hz)
        d = np.add(a, b, c)

        assert d is c
        assert c.sample_rate == 10 * u.Hz
        assert np.allclose(np.array(c), np.array(a + b))

    def test_multiple_outs(self):
        x = np.arange(16) / 4.0
        y, z = np.modf(x)

        a = pb.Signal(x, sample_rate=1 * u.Hz)
        b, c = np.modf(a)

        assert type(a) == type(b) == type(c)
        assert a.sample_rate == b.sample_rate == c.sample_rate
        assert np.allclose(b.data, y) and np.allclose(c.data, z)

        b = pb.Signal(np.zeros(x.shape), sample_rate=2 * u.Hz)
        c = pb.Signal(np.zeros(x.shape), sample_rate=5 * u.Hz)

        bb, cc = np.modf(a, b, c)

        assert bb is b and cc is c
        assert type(a) == type(bb) == type(cc)
        assert bb.sample_rate == 2 * u.Hz and cc.sample_rate == 5 * u.Hz
        assert np.allclose(bb.data, y) and np.allclose(cc.data, z)

        b = pb.Signal(np.zeros(x.shape), sample_rate=2 * u.Hz)
        c = pb.Signal(np.zeros(x.shape), sample_rate=5 * u.Hz)

        bb, cc = np.modf(a, b)

        assert bb is b and cc is not c
        assert type(a) == type(bb) == type(cc)
        assert bb.sample_rate == 2 * u.Hz and cc.sample_rate == a.sample_rate
        assert np.allclose(bb.data, y) and np.allclose(cc.data, z)

        b = pb.Signal(np.zeros(x.shape), sample_rate=2 * u.Hz)
        c = pb.Signal(np.zeros(x.shape), sample_rate=5 * u.Hz)

        bb, cc = np.modf(a, out=(None, c))

        assert bb is not b and cc is c
        assert type(a) == type(bb) == type(cc)
        assert bb.sample_rate == a.sample_rate and cc.sample_rate == 5 * u.Hz
        assert np.allclose(bb.data, y) and np.allclose(cc.data, z)


class TestSignalDaskFuncs:
    def test_compute(self):
        a = da.random.standard_normal((256, 16)).astype(np.float32)
        b = a.compute()

        x = pb.Signal(a, sample_rate=1 * u.Hz)

        for kw in [{}, {"scheduler": "threads"}]:
            y = x.compute(**kw)
            assert isinstance(y, type(x))
            assert isinstance(y.data, type(b))
            assert np.allclose(y.data, b)

            x = pb.Signal(b, sample_rate=1 * u.Hz)
            y = x.compute(**kw)
            assert isinstance(y, type(x))
            assert isinstance(y.data, type(b))
            assert np.allclose(y.data, b)

    def test_dask_persist(self):
        def bunch_of_operations(x):
            return np.abs(2 * (x - (x / 2))) + 1

        a = da.random.standard_normal((256, 16)).astype(np.float32)
        b = bunch_of_operations(a).compute()

        x = pb.Signal(a, sample_rate=1 * u.Hz)
        x = bunch_of_operations(x)

        for kw in [{}, {"scheduler": "threads"}]:
            y = x.persist(**kw)

            assert isinstance(y, pb.Signal)
            assert isinstance(y.data, type(a))

            assert len(x.data.__dask_graph__()) == 7
            assert len(y.data.__dask_graph__()) == 1
            assert np.allclose(y.data.compute(), b)

            z = pb.Signal(b, sample_rate=1 * u.Hz)
            y = z.persist(**kw)

            assert isinstance(y, pb.Signal)
            assert isinstance(y.data, type(b))
            assert np.allclose(y.data, b)

    def test_to_dask(self):
        a = RAND.standard_normal((256, 16), dtype=np.float32)
        x = pb.Signal(a, sample_rate=1 * u.Hz)
        y = x.to_dask_array()

        assert isinstance(y, pb.Signal)
        assert isinstance(y.data, da.Array)
        assert np.allclose(a, y.data.compute())

    def test_rechunk(self):
        a = RAND.standard_normal((256, 16), dtype=np.float32)
        x = pb.Signal(a, sample_rate=1 * u.Hz)
        y = x.rechunk((16, 4), balance=True)

        assert isinstance(y, pb.Signal)
        assert isinstance(y.data, da.Array)
        assert y.data.chunksize == (16, 4)
        assert np.allclose(a, y.data.compute())

        y = y.rechunk()

        assert isinstance(y, pb.Signal)
        assert isinstance(y.data, da.Array)
        assert y.data.chunksize[0] == 256
        assert np.allclose(a, y.data.compute())
